# coding=utf-8
# Copyright 2020 Microsoft and the Hugging Face Inc. team and Knowledgator.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Union
import torch
from torch import nn
import torch.nn.functional as F
from torch.nn.utils.rnn import pack_padded_sequence, pad_packed_sequence
from transformers.activations import ACT2FN

from .config import GLiClassModelConfig

class LstmSeq2SeqEncoder(nn.Module):
    def __init__(self, input_size, hidden_size, num_layers=1, dropout=0., bidirectional=False):
        super(LstmSeq2SeqEncoder, self).__init__()
        self.lstm = nn.LSTM(input_size=input_size,
                            hidden_size=hidden_size,
                            num_layers=num_layers,
                            dropout=dropout,
                            bidirectional=bidirectional,
                            batch_first=True)

    def forward(self, x, mask, hidden=None):
        # Packing the input sequence
        lengths = mask.sum(dim=1).cpu()
        packed_x = pack_padded_sequence(x, lengths, batch_first=True, enforce_sorted=False)

        # Passing packed sequence through LSTM
        packed_output, hidden = self.lstm(packed_x, hidden)

        # Unpacking the output sequence
        output, _ = pad_packed_sequence(packed_output, batch_first=True)

        return output

class FeaturesProjector(nn.Module):
    def __init__(self, config: GLiClassModelConfig):
        super().__init__()

        self.linear_1 = nn.Linear(config.encoder_config.hidden_size, config.hidden_size, bias=True)
        self.act = ACT2FN[config.projector_hidden_act]
        self.dropout = nn.Dropout(config.dropout)
        self.linear_2 = nn.Linear(config.hidden_size, config.encoder_config.hidden_size, bias=True)

    def forward(self, features):
        hidden_states = self.linear_1(features)
        hidden_states = self.act(hidden_states)
        hidden_states = self.dropout(hidden_states)
        hidden_states = self.linear_2(hidden_states)
        return hidden_states

class BiEncoderProjector(nn.Module):
    def __init__(self, config: GLiClassModelConfig):
        super().__init__()

        self.linear_1 = nn.Linear(config.label_model_config.hidden_size, config.hidden_size, bias=True)
        self.act = ACT2FN[config.projector_hidden_act]
        self.linear_2 = nn.Linear(config.hidden_size, config.encoder_config.hidden_size, bias=True)

    def forward(self, features):
        hidden_states = self.linear_1(features)
        hidden_states = self.act(hidden_states)
        hidden_states = self.linear_2(hidden_states)
        return hidden_states

# Copied from transformers.models.deberta.modeling_deberta.DropoutContext
class DropoutContext(object):
    def __init__(self):
        self.dropout = 0
        self.mask = None
        self.scale = 1
        self.reuse_mask = True


# Copied from transformers.models.deberta.modeling_deberta.get_mask
def get_mask(input, local_context):
    if not isinstance(local_context, DropoutContext):
        dropout = local_context
        mask = None
    else:
        dropout = local_context.dropout
        dropout *= local_context.scale
        mask = local_context.mask if local_context.reuse_mask else None

    if dropout > 0 and mask is None:
        mask = (1 - torch.empty_like(input).bernoulli_(1 - dropout)).to(torch.bool)

    if isinstance(local_context, DropoutContext):
        if local_context.mask is None:
            local_context.mask = mask

    return mask, dropout


# Copied from transformers.models.deberta.modeling_deberta.XDropout
class XDropout(torch.autograd.Function):
    """Optimized dropout function to save computation and memory by using mask operation instead of multiplication."""

    @staticmethod
    def forward(ctx, input, local_ctx):
        mask, dropout = get_mask(input, local_ctx)
        ctx.scale = 1.0 / (1 - dropout)
        if dropout > 0:
            ctx.save_for_backward(mask)
            return input.masked_fill(mask, 0) * ctx.scale
        else:
            return input

    @staticmethod
    def backward(ctx, grad_output):
        if ctx.scale > 1:
            (mask,) = ctx.saved_tensors
            return grad_output.masked_fill(mask, 0) * ctx.scale, None
        else:
            return grad_output, None

    @staticmethod
    def symbolic(g: torch._C.Graph, input: torch._C.Value, local_ctx: Union[float, DropoutContext]) -> torch._C.Value:
        from torch.onnx import symbolic_opset12

        dropout_p = local_ctx
        if isinstance(local_ctx, DropoutContext):
            dropout_p = local_ctx.dropout
        # StableDropout only calls this function when training.
        train = True
        # TODO: We should check if the opset_version being used to export
        # is > 12 here, but there's no good way to do that. As-is, if the
        # opset_version < 12, export will fail with a CheckerError.
        # Once https://github.com/pytorch/pytorch/issues/78391 is fixed, do something like:
        # if opset_version < 12:
        #   return torch.onnx.symbolic_opset9.dropout(g, input, dropout_p, train)
        return symbolic_opset12.dropout(g, input, dropout_p, train)
    
# Copied from transformers.models.deberta.modeling_deberta.StableDropout
class StableDropout(nn.Module):
    """
    Optimized dropout module for stabilizing the training

    Args:
        drop_prob (float): the dropout probabilities
    """

    def __init__(self, drop_prob):
        super().__init__()
        self.drop_prob = drop_prob
        self.count = 0
        self.context_stack = None

    def forward(self, x):
        """
        Call the module

        Args:
            x (`torch.tensor`): The input tensor to apply dropout
        """
        if self.training and self.drop_prob > 0:
            return XDropout.apply(x, self.get_context())
        return x

    def clear_context(self):
        self.count = 0
        self.context_stack = None

    def init_context(self, reuse_mask=True, scale=1):
        if self.context_stack is None:
            self.context_stack = []
        self.count = 0
        for c in self.context_stack:
            c.reuse_mask = reuse_mask
            c.scale = scale

    def get_context(self):
        if self.context_stack is not None:
            if self.count >= len(self.context_stack):
                self.context_stack.append(DropoutContext())
            ctx = self.context_stack[self.count]
            ctx.dropout = self.drop_prob
            self.count += 1
            return ctx
        else:
            return self.drop_prob

class SelfAttentionBlock(nn.Module):
    def __init__(self, d_model, num_heads, dropout=0.1):
        super().__init__()
        self.self_attn = nn.MultiheadAttention(d_model, num_heads, dropout=dropout)
        self.norm = nn.LayerNorm(d_model)
        self.dropout = nn.Dropout(dropout)

    def forward(self, x, mask=None):
        attn_output, _ = self.self_attn(x, x, x, attn_mask=mask)
        return self.norm(x + self.dropout(attn_output))

class CrossAttentionBlock(nn.Module):
    def __init__(self, d_model, num_heads, dropout=0.1):
        super().__init__()
        self.cross_attn = nn.MultiheadAttention(d_model, num_heads, dropout=dropout)
        self.norm = nn.LayerNorm(d_model)
        self.dropout = nn.Dropout(dropout)

    def forward(self, query, key, value, mask=None):
        attn_output, _ = self.cross_attn(query, key, value, attn_mask=mask)
        return self.norm(query + self.dropout(attn_output))

class Fuser(nn.Module):
    def __init__(self, d_model, num_heads, num_layers, dropout=0.1):
        super().__init__()
        self.d_model = d_model
        self.layers = nn.ModuleList([
            nn.ModuleList([
                SelfAttentionBlock(d_model, num_heads, dropout),
                CrossAttentionBlock(d_model, num_heads, dropout)
            ])
            for _ in range(num_layers)
        ])
        self.fc = nn.Linear(d_model, d_model)

    def forward(self, query, key, query_mask=None, key_mask=None):
        if query_mask is not None and key_mask is not None:
            self_attn_mask = query_mask.unsqueeze(1) * query_mask.unsqueeze(2)
            cross_attn_mask = query_mask.unsqueeze(-1) * key_mask.unsqueeze(1)
        else:
            self_attn_mask = None
            cross_attn_mask = None

        value = self.fc(key)

        for self_attn, cross_attn in self.layers:
            query = self_attn(query, mask=self_attn_mask)
            query = cross_attn(query, key, value, mask=cross_attn_mask)

        return query

class LayerwiseAttention(nn.Module):
    def __init__(self, num_layers, hidden_size, output_size=None):
        super().__init__()
        self.num_layers = num_layers
        self.hidden_size = hidden_size
        self.output_size = output_size if output_size is not None else hidden_size
        
        # Squeeze operation
        self.squeeze = nn.Linear(hidden_size, 1)
        
        # Excitation operation
        self.W1 = nn.Linear(num_layers, num_layers // 2)
        self.W2 = nn.Linear(num_layers // 2, num_layers)
        
        # Final projection
        self.output_projection = nn.Linear(self.hidden_size, self.output_size)
        
    def forward(self, encoder_outputs):
        # encoder_outputs is a list of tensors, each of shape [B, L, D]
        B, L, D = encoder_outputs[0].shape
        
        # Concatenate all layers
        U = torch.stack(encoder_outputs, dim=1)  # [B, K, L, D]
        
        # Squeeze operation
        Z = self.squeeze(U).squeeze(-1)  # [B, K, L]
        Z = Z.mean(dim=2)  # [B, K]
        
        # Excitation operation
        s = self.W2(F.relu(self.W1(Z)))  # [B, K]
        s = torch.sigmoid(s)  # [B, K]
        
        # Apply attention weights
        U_weighted = U * s.unsqueeze(-1).unsqueeze(-1)  # [B, K, L, D]
        
        # Sum across layers
        U_sum = U_weighted.sum(dim=1)  # [B, L, D]
        
        # Final projection
        output = self.output_projection(U_sum)  # [B, L, output_size]
        
        return output


import torch

def is_module_available(module_name):
    """
    Checks whether the specified Python module is available.
    
    Args:
        module_name (str): The name of the module to check.
        
    Returns:
        bool: True if the module is available, False otherwise.
    """
    try:
        __import__(module_name)
        return True
    except ImportError:
        return False

class MissedPackageException(Exception):
    """Raised when the requested decoder model is not supported."""
    pass


def retrieval_augmented_text(text: str, examples: list) -> str:
    """
    Constructs a new text by appending relevant retrieved examples to the input text.

    Args:
        text (str): The main input text.
        examples (list): A list of examples in the format
                         {"text": str, "true_labels": List[str], "all_labels": List[str]}.

    Returns:
        str: The modified text with relevant examples appended.
    """
    if not examples:
        return text

    retrieved_examples = []
    all_labels = set(label for example in examples for label in example.get("true_labels", []))
    relevant_examples = [ex for ex in examples if set(ex.get("true_labels", [])) & all_labels]

    for example in relevant_examples:
        example_text = example["text"]
        true_labels = example.get("true_labels", [])
        all_labels = example.get("all_labels", [])

        false_labels = list(set(all_labels) - set(true_labels))

        true_labels_str = " ".join([f"<<TRUE_LABEL>> {label}" for label in true_labels])
        false_labels_str = " ".join([f"<<FALSE_LABEL>> {label}" for label in false_labels])

        retrieved_example_str = f"<<EXAMPLE>> {example_text} {true_labels_str} {false_labels_str} <</EXAMPLE>>"
        retrieved_examples.append(retrieved_example_str)

    augmented_text = f"{text} {' '.join(retrieved_examples)}" if retrieved_examples else text

    return augmented_text

def default_f1_reward(
    probs: torch.Tensor,
    actions: torch.Tensor,
    original_targets: torch.Tensor,
    valid_mask: torch.Tensor
) -> torch.Tensor:
    """
    A variant that extracts list-of-indices sets and then calculates
    the F1 score in a classical manner. Returns shape (N, 1).
    
    Args:
        probs:              (N, T) Tensor of probabilities (not used here but left for interface consistency).
        actions:            (N, T) Tensor of predicted labels in {0, 1}.
        original_targets:   (N, T) Tensor of ground-truth labels in {0, 1}.
        valid_mask:         (N, T) Tensor indicating which positions are valid (1) vs. invalid (0).

    Returns:
        f1_scores: (N, 1) Tensor containing the F1 score for each row.
    """
    N = actions.shape[0]
    f1_scores = []

    for i in range(N):
        # Filter valid positions
        valid_preds_i = actions[i] * valid_mask[i]
        valid_targets_i = original_targets[i] * valid_mask[i]

        # Get the set of indices where we predicted 1
        predicted_set = set((valid_preds_i == 1).nonzero(as_tuple=True)[0].tolist())
        # Get the set of indices where the ground truth is 1
        target_set = set((valid_targets_i == 1).nonzero(as_tuple=True)[0].tolist())

        # Compute intersection
        intersection = predicted_set.intersection(target_set)

        # Precision
        if len(predicted_set) > 0:
            precision = len(intersection) / len(predicted_set)
        else:
            precision = 0.0

        # Recall
        if len(target_set) > 0:
            recall = len(intersection) / len(target_set)
        else:
            recall = 0.0

        # F1 score
        if (precision + recall) > 0:
            f1 = 2 * precision * recall / (precision + recall)
        else:
            f1 = 0.0

        f1_scores.append(f1)
    
    # Convert list to tensor shape (N, 1)
    f1_scores = torch.tensor(f1_scores, dtype=torch.float).unsqueeze(-1)
    return f1_scores.detach().to(probs.device)


import torch
from torch import nn

class ScorerWeightedDot(nn.Module):
    def __init__(self, hidden_size, dropout=0.1):
        super().__init__()

        self.proj_text = nn.Linear(hidden_size, hidden_size * 2)
        self.proj_label = nn.Linear(hidden_size, hidden_size * 2)

        self.out_mlp = nn.Sequential(
            nn.Linear(hidden_size * 3, hidden_size * 4),
            nn.Dropout(dropout),
            nn.ReLU(),
            nn.Linear(hidden_size * 4, 1)  # start, end, score
        )

    def forward(self, text_rep, label_rep):
        batch_size, hidden_size = text_rep.shape
        num_classes = label_rep.shape[1]

        # (batch_size, 1, 3, hidden_size)
        text_rep = self.proj_text(text_rep).view(batch_size, 1, 1, 2, hidden_size)
        label_rep = self.proj_label(label_rep).view(batch_size, 1, num_classes, 2, hidden_size)

        # (2, batch_size, 1, num_classes, hidden_size)
        text_rep = text_rep.expand(-1, -1, num_classes, -1, -1).permute(3, 0, 1, 2, 4)
        label_rep = label_rep.expand(-1, 1, -1, -1, -1).permute(3, 0, 1, 2, 4)

        # (batch_size, 1, num_classes, hidden_size * 3)
        cat = torch.cat([text_rep[0], label_rep[0], text_rep[1] * label_rep[1]], dim=-1)

        # (batch_size, num_classes)
        scores = self.out_mlp(cat).view(batch_size, num_classes)

        return scores
    
class ScorerDot(nn.Module):
    def __init__(self, *args):
        super().__init__()
        pass

    def forward(self, text_rep, label_rep):
        # dot product with einsum
        scores = torch.einsum('BD,BCD->BC', text_rep, label_rep)
        return scores
    
class MLPScorer(nn.Module):
    def __init__(self, hidden_size, mlp_hidden_size=256):
        super().__init__()
        
        # Calculate the input size for the MLP
        total_input_size = hidden_size*2
        
        # Define the MLP
        self.mlp = nn.Sequential(
            nn.Linear(total_input_size, mlp_hidden_size),
            nn.ReLU(),
            nn.Linear(mlp_hidden_size, mlp_hidden_size // 2),
            nn.ReLU(),
            nn.Linear(mlp_hidden_size // 2, 1)
        )

    def forward(self, text_rep, label_rep):
        # Concatenate text and label representations
        batch_size, num_labels, dim = label_rep.shape
        text_rep = text_rep.unsqueeze(1).expand(batch_size, num_labels, dim)
        combined_rep = torch.cat([text_rep, label_rep], dim=-1)
        
        # Pass through MLP
        scores = self.mlp(combined_rep).squeeze(-1)
        
        return scores

class HopfieldScorer(nn.Module):
    def __init__(self, hidden_size, mlp_hidden_size=256, beta=4, num_iteration=1):
        super().__init__()
        
        self.q_proj = nn.Linear(hidden_size, hidden_size, bias=False)
        self.k_proj = nn.Linear(hidden_size, hidden_size, bias=False)
        self.v_proj = nn.Linear(hidden_size, hidden_size, bias=False)
        
        # Define the MLP
        self.mlp = nn.Sequential(
            nn.Linear(hidden_size, mlp_hidden_size),
            nn.ReLU(),
            nn.Linear(mlp_hidden_size, mlp_hidden_size // 2),
            nn.ReLU(),
            nn.Linear(mlp_hidden_size // 2, 1)
        )

        self.beta = beta
        self.num_iteration = num_iteration

    def forward(self, text_rep, label_rep):
        """
        text_rep: [batch_size, hidden_size]
        label_rep: [batch_size, num_labels, hidden_size]
        """        
        for i in range(self.num_iteration):
            # Expand text_rep to match label_rep’s batch shape
            text_rep_expanded = text_rep.unsqueeze(1)  # [batch_size, 1, dim]

            # Compute Q, K, V
            query = self.q_proj(label_rep)           # [batch_size, num_labels, dim]
            key   = self.k_proj(text_rep_expanded)   # [batch_size, 1, dim]
            value = self.v_proj(text_rep_expanded)   # [batch_size, 1, dim]


            attn = torch.bmm(query, key.transpose(1, 2))  # [b, num_labels, 1]
            attn = attn * self.beta                       # optional beta scaling
            attn = torch.nn.functional.softmax(attn, dim=1)                 # softmax over labels

            context = attn * value  # [b, num_labels, dim]

            label_rep = label_rep + context

        scores = self.mlp(label_rep).squeeze(-1)  # [b, num_labels]
        
        return scores


# Example dictionary for scorers
SCORER2OBJECT = {
    "weighted-dot": ScorerWeightedDot, 
    "simple": ScorerDot, 
    "mlp": MLPScorer,
    "hopfield": HopfieldScorer  # <-- Add reference here if you want
}


from typing import Optional

import torch
from torch import nn


class GlobalMaxPooling1D(nn.Module):
    """Applies Global Max Pooling on the timesteps dimension."""

    def forward(self, x: torch.Tensor):
        return x.amax(dim=1)


class FirstTokenPooling1D(nn.Module):
    """Takes the first token's embedding."""

    def forward(self, x: torch.Tensor):
        return x[:, 0, :]


class LastTokenPooling1D(nn.Module):
    """Takes the last token's embedding."""

    def forward(self, x: torch.Tensor):
        return x[:, -1, :]


class GlobalAvgPooling1D(nn.Module):
    """Applies Global Average Pooling on the timesteps dimension."""

    def forward(
        self, x: torch.Tensor, attention_mask: Optional[torch.Tensor] = None
    ):
        if attention_mask is not None:
            attention_mask = attention_mask.repeat((1, 1, x.shape[-1])).to(
                dtype=x.dtype
            )
            x = x * attention_mask
            return x.sum(1) / attention_mask.sum(1)
        else:
            return x.mean(dim=1)


class GlobalSumPooling1D(nn.Module):
    """Applies Global Sum Pooling on the timesteps dimension."""

    def forward(self, x: torch.Tensor, attention_mask: Optional[torch.Tensor] = None):
        if attention_mask is not None:
            x = x * attention_mask
        return x.sum(dim=1)


class GlobalRMSPooling1D(nn.Module):
    """Applies Global RMS Pooling on the timesteps dimension."""

    def forward(self, x: torch.Tensor, attention_mask: Optional[torch.Tensor] = None):
        if attention_mask is not None:
            attention_mask = attention_mask.repeat((1, 1, x.shape[-1])).to(
                dtype=x.dtype
            )
            x = x * attention_mask
            return (x.pow(2).sum(dim=1) / attention_mask.sum(1)).sqrt()
        else:
            return x.pow(2).mean(dim=1).sqrt()


class GlobalAbsMaxPooling1D(nn.Module):
    """Applies Global Max Pooling of absolute values on the timesteps dimension."""

    def forward(self, x: torch.Tensor, attention_mask: Optional[torch.Tensor] = None):
        if attention_mask is not None:
            attention_mask = attention_mask.repeat((1, 1, x.shape[-1])).to(
                dtype=x.dtype
            )
            x = x * attention_mask
        return x.abs().amax(dim=1)


class GlobalAbsAvgPooling1D(nn.Module):
    """Applies Global Average Pooling of absolute values on the timesteps dimension."""

    def forward(self, x: torch.Tensor, attention_mask: Optional[torch.Tensor] = None):
        if attention_mask is not None:
            attention_mask = attention_mask.repeat((1, 1, x.shape[-1])).to(
                dtype=x.dtype
            )
            x = (x * attention_mask).abs()
            return x.sum(dim=1) / attention_mask.sum(1)
        else:
            return x.abs().mean(dim=1)

POOLING2OBJECT = {
    'max': GlobalMaxPooling1D,
    'first': FirstTokenPooling1D,
    'last': LastTokenPooling1D,
    'avg': GlobalAvgPooling1D,
    'sum': GlobalSumPooling1D,
    'rms': GlobalRMSPooling1D,
    'abs_max': GlobalAbsMaxPooling1D,
    'abs_avg': GlobalAbsAvgPooling1D
}


import torch
from tqdm import tqdm
from typing import List, Dict, Union, Optional, Any, Tuple
from transformers import AutoTokenizer
from abc import ABC, abstractmethod
from .model import GLiClassModel, GLiClassBiEncoder
from .utils import retrieval_augmented_text


def flatten_hierarchical_labels(
    labels: Union[List[str], Dict[str, Any]],
    prefix: str = "",
    separator: str = "."
) -> List[str]:
    """
    Flatten hierarchical labels into dot notation.
    
    Supports arbitrary nesting depth. Examples:
    
    Input: {"sentiment": ["positive", "negative", "neutral"], "topic": ["product", "service", "shipping"]}
    Output: ["sentiment.positive", "sentiment.negative", "sentiment.neutral", 
             "topic.product", "topic.service", "topic.shipping"]
    
    Input: {
        "category": {
            "electronics": ["phone", "laptop"],
            "clothing": ["shirt", "pants"]
        }
    }
    Output: [
        "category.electronics.phone",
        "category.electronics.laptop", 
        "category.clothing.shirt",
        "category.clothing.pants"
    ]
    
    Input: ["label1", "label2"]  # Already flat
    Output: ["label1", "label2"]
    
    Args:
        labels: Either a list of string labels or a hierarchical dict
        prefix: Current prefix for recursion (internal use)
        separator: Separator to use between hierarchy levels (default: ".")
        
    Returns:
        List of flattened label strings with dot notation
    """
    if isinstance(labels, list):
        if prefix:
            return [f"{prefix}{separator}{label}" for label in labels]
        return labels
    
    elif isinstance(labels, dict):
        flattened = []
        for key, value in labels.items():
            new_prefix = f"{prefix}{separator}{key}" if prefix else key
            flattened.extend(flatten_hierarchical_labels(value, new_prefix, separator))
        return flattened
    
    elif isinstance(labels, str):
        if prefix:
            return [f"{prefix}{separator}{labels}"]
        return [labels]
    
    else:
        raise ValueError(f"Unsupported label type: {type(labels)}. Expected list, dict, or str.")


def build_hierarchical_output(
    predictions: List[Dict[str, float]],
    original_labels: Union[List[str], Dict[str, Any]],
    separator: str = ".",
    all_scores: Optional[Dict[str, float]] = None
) -> Union[Dict[str, float], Dict[str, Any]]:
    """
    Build hierarchical output structure matching the input labels structure.
    
    Args:
        predictions: List of prediction dicts with 'label' and 'score'
        original_labels: Original hierarchical labels structure
        separator: Separator used in flattened labels
        all_scores: Optional dict of all label scores (for complete output)
        
    Returns:
        Hierarchical structure with scores matching the input format
        
    Example:
        Input predictions: [
            {'label': 'sentiment.positive', 'score': 0.85},
            {'label': 'topic.product', 'score': 0.72}
        ]
        Input original_labels: {
            "sentiment": ["positive", "negative", "neutral"],
            "topic": ["product", "service", "shipping"]
        }
        Output: {
            "sentiment": {"positive": 0.85, "negative": 0.0, "neutral": 0.0},
            "topic": {"product": 0.72, "service": 0.0, "shipping": 0.0}
        }
    """
    score_lookup = {pred['label']: pred['score'] for pred in predictions}
    
    if all_scores:
        for k, v in all_scores.items():
            if k not in score_lookup:
                score_lookup[k] = v
    
    def _build_recursive(
        structure: Union[List[str], Dict[str, Any]], 
        prefix: str = ""
    ) -> Union[Dict[str, float], Dict[str, Any]]:
        if isinstance(structure, list):
            result = {}
            for label in structure:
                full_label = f"{prefix}{separator}{label}" if prefix else label
                result[label] = score_lookup.get(full_label, 0.0)
            return result
        
        elif isinstance(structure, dict):
            result = {}
            for key, value in structure.items():
                new_prefix = f"{prefix}{separator}{key}" if prefix else key
                result[key] = _build_recursive(value, new_prefix)
            return result
        
        elif isinstance(structure, str):
            full_label = f"{prefix}{separator}{structure}" if prefix else structure
            return {structure: score_lookup.get(full_label, 0.0)}
        
        return {}
    
    if isinstance(original_labels, list):
        return {label: score_lookup.get(label, 0.0) for label in original_labels}
    
    return _build_recursive(original_labels)


def format_examples_prompt(
    examples: List[Dict[str, Any]],
    example_token: str = "<<EXAMPLE>>",
    sep_token: str = "<<SEP>>"
) -> str:
    """
    Format few-shot examples into a prompt string using <<EXAMPLE>> token.

    Format matches training: <<EXAMPLE>>text \nLabels:\n label1, label2
    with a single <<SEP>> after all examples.

    Args:
        examples: List of example dicts with 'text' and 'labels'/'true_labels' keys
        example_token: Token to mark examples (default: "<<EXAMPLE>>")
        sep_token: Separator token after all examples (default: "<<SEP>>")

    Returns:
        Formatted examples string
    """
    if not examples:
        return ""

    formatted_parts = []
    for example in examples:
        text = example.get('text', '')
        labels = example.get('labels', example.get('true_labels', []))

        if isinstance(labels, list):
            labels_str = ', '.join(labels)
        else:
            labels_str = str(labels)

        # Match training format: " \nLabels:\n " instead of "\nLabels: "
        formatted_parts.append(f"{example_token}{text} \nLabels:\n {labels_str}")

    # Add single SEP token after all examples (matching training)
    formatted_parts.append(sep_token)

    return ''.join(formatted_parts)


class BaseZeroShotClassificationPipeline(ABC):
    def __init__(self, model, tokenizer, max_classes=25, max_length=1024, 
                 classification_type='multi-label', device='cuda:0', progress_bar=True,
                 label_separator: str = "."):
        self.model = model
        if isinstance(tokenizer, str):
            self.tokenizer = AutoTokenizer.from_pretrained(tokenizer)
        else:
            self.tokenizer = tokenizer
        self.max_classes = max_classes
        self.classification_type = classification_type
        self.max_length = max_length
        self.progress_bar = progress_bar
        self.label_separator = label_separator
        
        self.example_token = "<<EXAMPLE>>"
        self.label_token = "<<LABEL>>"
        self.sep_token = "<<SEP>>"

        if not isinstance(device, torch.device):
            if torch.cuda.is_available() and 'cuda' in device:
                self.device = torch.device(device)
            else:
                self.device = torch.device('cpu')
        else:
            self.device = device

        if self.model.device != self.device:
            self.model.to(self.device)

        # Ensure model is in evaluation mode for inference
        self.model.eval()

    def _process_labels(
        self, 
        labels: Union[List[str], Dict[str, Any], List[List[str]], List[Dict[str, Any]]]
    ) -> Union[List[str], List[List[str]]]:
        """Process labels to handle hierarchical structures."""
        if not labels:
            return labels
        
        if isinstance(labels, dict):
            return flatten_hierarchical_labels(labels, separator=self.label_separator)
        
        if isinstance(labels, list):
            if len(labels) == 0:
                return labels
            
            first_elem = labels[0]
            
            if isinstance(first_elem, str):
                return labels
            
            if isinstance(first_elem, dict):
                return [
                    flatten_hierarchical_labels(lbl, separator=self.label_separator) 
                    for lbl in labels
                ]
            
            if isinstance(first_elem, list):
                if first_elem and isinstance(first_elem[0], dict):
                    return [
                        flatten_hierarchical_labels(lbl, separator=self.label_separator) 
                        for lbl in labels
                    ]
                return labels
        
        return labels

    def _format_examples_for_input(
        self, 
        examples: Optional[List[Dict[str, Any]]] = None
    ) -> str:
        """Format few-shot examples using <<EXAMPLE>> and <<SEP>> tokens."""
        if not examples:
            return ""
        return format_examples_prompt(
            examples, 
            example_token=self.example_token,
            sep_token=self.sep_token
        )
    
    def _format_prompt(
        self,
        prompt: Optional[Union[str, List[str]]] = None,
        index: int = 0
    ) -> str:
        """Format the task description prompt."""
        if prompt is None:
            return ""
        
        if isinstance(prompt, str):
            return prompt
        
        if isinstance(prompt, list):
            if index < len(prompt):
                return prompt[index]
            return prompt[0] if prompt else ""
        
        return ""

    @abstractmethod
    def prepare_inputs(self, texts, labels, same_labels=False, examples=None, prompt=None):
        pass
    
    def _get_batch_examples(self, examples, start_idx, batch_size):
        """Get examples for current batch."""
        if not examples:
            return None
        if isinstance(examples[0], list):
            return examples[start_idx:start_idx + batch_size]
        return examples
    
    def _get_batch_prompt(self, prompt, start_idx, batch_size):
        """Get prompt for current batch."""
        if not prompt:
            return None
        if isinstance(prompt, list):
            return prompt[start_idx:start_idx + batch_size]
        return prompt
    
    @torch.no_grad()
    def get_embeddings(self, texts, labels, batch_size=8, examples=None, prompt=None):
        if isinstance(texts, str):
            texts = [texts]
        
        labels = self._process_labels(labels)
        
        if isinstance(labels[0], str):
            same_labels = True
        else:
            same_labels = False
        
        results = []

        iterable = range(0, len(texts), batch_size)
        if self.progress_bar:
            iterable = tqdm(iterable)

        for idx in iterable:
            batch_texts = texts[idx:idx+batch_size]
            batch_examples = self._get_batch_examples(examples, idx, len(batch_texts))
            batch_prompt = self._get_batch_prompt(prompt, idx, len(batch_texts))
            
            tokenized_inputs = self.prepare_inputs(
                batch_texts, labels, same_labels, 
                examples=batch_examples, prompt=batch_prompt
            )
            model_output = self.model(
                **tokenized_inputs, 
                output_text_embeddings=True,
                output_class_embeddings=True
            )
            logits = model_output.logits
            text_embeddings = model_output.text_embeddings
            class_embeddings = model_output.class_embeddings
            batch_size_actual = logits.shape[0]
            
            for i in range(batch_size_actual):
                result = {
                    'logits': logits[i].cpu().numpy(),
                    'text_embedding': text_embeddings[i].cpu().numpy(),
                    'class_embeddings': class_embeddings[i].cpu().numpy()
                }
                results.append(result)
        
        return results

    @torch.no_grad()
    def __call__(
        self, 
        texts: Union[str, List[str]], 
        labels: Union[List[str], Dict[str, Any], List[List[str]], List[Dict[str, Any]]],
        threshold: float = 0.5, 
        batch_size: int = 8, 
        rac_examples: Optional[List] = None,
        examples: Optional[List[Dict[str, Any]]] = None,
        prompt: Optional[Union[str, List[str]]] = None,
        return_hierarchical: bool = False
    ):
        """
        Perform zero-shot classification.
        
        Args:
            texts: Single text or list of texts to classify
            labels: Labels in various formats (flat list or hierarchical dict)
            threshold: Classification threshold for multi-label (default: 0.5)
            batch_size: Batch size for processing
            rac_examples: Retrieval augmented examples (legacy)
            examples: Few-shot examples with 'text' and 'labels'/'true_labels' keys
            prompt: Task description - string (same for all) or list (per-text)
            return_hierarchical: If True, return hierarchical structure with all scores
                
        Returns:
            List of classification results or hierarchical dict structure.
        """
        original_labels = labels
        
        if isinstance(texts, str):
            if rac_examples:
                texts = retrieval_augmented_text(texts, rac_examples)
            texts = [texts]
        else:
            if rac_examples:
                texts = [
                    retrieval_augmented_text(text, ex) 
                    for text, ex in zip(texts, rac_examples)
                ]
        
        processed_labels = self._process_labels(labels)
        
        if isinstance(processed_labels[0], str):
            same_labels = True
        else:
            same_labels = False
            
        results = []
        all_scores_list = []
        
        iterable = range(0, len(texts), batch_size)
        if self.progress_bar:
            iterable = tqdm(iterable)

        for idx in iterable:
            batch_texts = texts[idx:idx+batch_size]
            if not same_labels:
                batch_labels = processed_labels[idx:idx+batch_size]
            else:
                batch_labels = processed_labels
            
            batch_examples = self._get_batch_examples(examples, idx, len(batch_texts))
            batch_prompt = self._get_batch_prompt(prompt, idx, len(batch_texts))
            
            tokenized_inputs = self.prepare_inputs(
                batch_texts, batch_labels, same_labels, 
                examples=batch_examples, prompt=batch_prompt
            )
            model_output = self.model(**tokenized_inputs)
            logits = model_output.logits
            
            if self.classification_type == 'single-label':
                for i in range(len(batch_texts)):
                    score = torch.softmax(logits[i], dim=-1)
                    if same_labels:
                        curr_labels = batch_labels
                    else:
                        curr_labels = batch_labels[i]
                    
                    if return_hierarchical:
                        all_scores = {curr_labels[j]: score[j].item() for j in range(len(curr_labels))}
                        all_scores_list.append(all_scores)
                    
                    pred_label = curr_labels[torch.argmax(score).item()]
                    results.append([{'label': pred_label, 'score': score.max().item()}])
                    
            elif self.classification_type == 'multi-label':
                sigmoid = torch.nn.Sigmoid()
                probs = sigmoid(logits)
                for i in range(len(batch_texts)):
                    text_results = []
                    if same_labels:
                        curr_labels = batch_labels
                    else:
                        curr_labels = batch_labels[i]
                    
                    if return_hierarchical:
                        all_scores = {curr_labels[j]: probs[i][j].item() for j in range(len(curr_labels))}
                        all_scores_list.append(all_scores)
                    
                    for j, prob in enumerate(probs[i][:len(curr_labels)]):
                        score = prob.item()
                        if score >= threshold:
                            text_results.append({'label': curr_labels[j], 'score': score})
                    results.append(text_results)
            else:
                raise ValueError(
                    "Unsupported classification type: choose 'single-label' or 'multi-label'"
                )
        
        if return_hierarchical:
            hierarchical_results = []
            for i, (result, all_scores) in enumerate(zip(results, all_scores_list)):
                if same_labels:
                    orig_lbl = original_labels
                else:
                    orig_lbl = original_labels[i] if i < len(original_labels) else original_labels
                
                hierarchical_results.append(
                    build_hierarchical_output(result, orig_lbl, self.label_separator, all_scores)
                )
            return hierarchical_results
        
        return results


class UniEncoderZeroShotClassificationPipeline(BaseZeroShotClassificationPipeline):
    def __init__(self, model, tokenizer, max_classes=25, max_length=1024, 
                 classification_type='multi-label', device='cuda:0', progress_bar=True,
                 label_separator: str = "."):
        super().__init__(
            model, tokenizer, max_classes, max_length, 
            classification_type, device, progress_bar, label_separator
        )

    def prepare_input(self, text, labels, examples=None, prompt=None):
        """
        Prepare input matching training format from data_processing.py:
        Order: Labels → SEP → Prompt → Text → Examples
        """
        input_parts = []

        # 1. Add labels
        for label in labels:
            label_tag = f"{self.label_token}{label}"
            input_parts.append(label_tag)
        input_parts.append(self.sep_token)

        # 2. Add task description prompt
        if prompt:
            input_parts.append(prompt)

        # 3. Format examples to go after text
        examples_str = ""
        if examples:
            examples_str = self._format_examples_for_input(examples)

        if self.model.config.prompt_first:
            return ''.join(input_parts) + text + examples_str
        else:
            return text + ''.join(input_parts) + examples_str

    def prepare_inputs(self, texts, labels, same_labels=False, examples=None, prompt=None):
        inputs = []

        if same_labels:
            for i, text in enumerate(texts):
                text_examples = None
                if examples:
                    if isinstance(examples[0], list):
                        text_examples = examples[i] if i < len(examples) else None
                    else:
                        text_examples = examples

                text_prompt = self._format_prompt(prompt, i)
                inputs.append(self.prepare_input(text, labels, text_examples, text_prompt))
        else:
            for i, (text, labels_) in enumerate(zip(texts, labels)):
                text_examples = None
                if examples:
                    if isinstance(examples[0], list):
                        text_examples = examples[i] if i < len(examples) else None
                    else:
                        text_examples = examples

                text_prompt = self._format_prompt(prompt, i)
                inputs.append(self.prepare_input(text, labels_, text_examples, text_prompt))

        tokenized_inputs = self.tokenizer(
            inputs, truncation=True,
            max_length=self.max_length,
            padding="longest", return_tensors="pt"
        ).to(self.device)

        return tokenized_inputs


class EncoderDecoderZeroShotClassificationPipeline(BaseZeroShotClassificationPipeline):
    def __init__(self, model, tokenizer, max_classes=25, max_length=1024,
                 classification_type='multi-label', device='cuda:0', progress_bar=True,
                 label_separator: str = "."):
        super().__init__(
            model, tokenizer, max_classes, max_length,
            classification_type, device, progress_bar, label_separator
        )

    def prepare_labels_prompt(self, labels, prompt=None):
        """Match training format: Labels → SEP → Prompt"""
        input_parts = []

        for label in labels:
            # label_tag = f"{label}{self.label_token}"
            label_tag = f"{self.label_token}{label}"
            input_parts.append(label_tag)
        input_parts.append(self.sep_token)

        if prompt:
            input_parts.append(prompt)

        return ''.join(input_parts)

    def prepare_inputs(self, texts, labels, same_labels=False, examples=None, prompt=None):
        prompts = []
        processed_texts = []

        if same_labels:
            for i, text in enumerate(texts):
                text_examples = None
                if examples:
                    if isinstance(examples[0], list):
                        text_examples = examples[i] if i < len(examples) else None
                    else:
                        text_examples = examples
                text_prompt = self._format_prompt(prompt, i)
                prompts.append(self.prepare_labels_prompt(labels, text_prompt))
                examples_str = self._format_examples_for_input(text_examples) if text_examples else ""
                processed_texts.append(text + examples_str)
        else:
            for i, labels_ in enumerate(labels):
                text_examples = None
                if examples:
                    if isinstance(examples[0], list):
                        text_examples = examples[i] if i < len(examples) else None
                    else:
                        text_examples = examples
                text_prompt = self._format_prompt(prompt, i)
                prompts.append(self.prepare_labels_prompt(labels_, text_prompt))
                examples_str = self._format_examples_for_input(text_examples) if text_examples else ""
                processed_texts.append(texts[i] + examples_str)
        
        tokenized_inputs = self.tokenizer(
            processed_texts, truncation=True, 
            max_length=self.max_length, 
            padding="longest", return_tensors="pt"
        ).to(self.device)
        
        tokenized_classes = self.tokenizer(
            prompts, max_length=self.max_length, 
            truncation=True, padding="longest", return_tensors='pt'
        ).to(self.device)
        
        tokenized_inputs["class_input_ids"] = tokenized_classes["input_ids"]
        tokenized_inputs["class_attention_mask"] = tokenized_classes["attention_mask"]

        return tokenized_inputs

    
class BiEncoderZeroShotClassificationPipeline(BaseZeroShotClassificationPipeline):
    def __init__(self, model, tokenizer, max_classes=25, max_length=1024, 
                 classification_type='multi-label', device='cuda:0', progress_bar=True,
                 label_separator: str = "."):
        super().__init__(
            model, tokenizer, max_classes, max_length, 
            classification_type, device, progress_bar, label_separator
        )
        self.labels_tokenizer = AutoTokenizer.from_pretrained(model.config.label_model_name)

    def prepare_input(self, text, labels, examples=None, prompt=None):
        input_parts = []

        if prompt:
            input_parts.append(prompt)
            input_parts.append(" ")

        for label in labels:
            input_parts.append(self.label_token)
        input_parts.append(self.sep_token)

        examples_str = ""
        if examples:
            examples_str = self._format_examples_for_input(examples)

        if self.model.config.prompt_first:
            return ''.join(input_parts) + text + examples_str
        else:
            return text + ''.join(input_parts) + examples_str
    
    def prepare_inputs(self, texts, labels, same_labels=False, examples=None, prompt=None):
        if self.model.config.architecture_type == 'bi-encoder-fused':
            inputs = []
            if same_labels:
                for i, text in enumerate(texts):
                    text_examples = None
                    if examples:
                        if isinstance(examples[0], list):
                            text_examples = examples[i] if i < len(examples) else None
                        else:
                            text_examples = examples
                    text_prompt = self._format_prompt(prompt, i)
                    inputs.append(self.prepare_input(text, labels, text_examples, text_prompt))
            else:
                for i, (text, labels_) in enumerate(zip(texts, labels)):
                    text_examples = None
                    if examples:
                        if isinstance(examples[0], list):
                            text_examples = examples[i] if i < len(examples) else None
                        else:
                            text_examples = examples
                    text_prompt = self._format_prompt(prompt, i)
                    inputs.append(self.prepare_input(text, labels_, text_examples, text_prompt))
        else:
            inputs = []
            for i, text in enumerate(texts):
                text_prompt = self._format_prompt(prompt, i)
                if text_prompt:
                    inputs.append(f"{text_prompt} {text}")
                else:
                    inputs.append(text)
            
        if same_labels:
            tokenized_inputs = self.tokenizer(
                inputs, truncation=True,
                max_length=self.max_length,
                padding="longest", return_tensors="pt"
            ).to(self.device)

            tokenized_labels = self.labels_tokenizer(
                labels, truncation=True,
                max_length=self.max_length,
                padding="longest", return_tensors="pt"
            ).to(self.device)
            
            tokenized_inputs['class_input_ids'] = tokenized_labels['input_ids'].expand(
                len(texts), -1, -1
            )
            tokenized_inputs['class_attention_mask'] = tokenized_labels['attention_mask'].expand(
                len(texts), -1, -1
            )
            
            labels_mask = [[1 for _ in range(len(labels))] for _ in range(len(texts))]
            tokenized_inputs["labels_mask"] = torch.tensor(labels_mask).to(self.device)
        else:
            tokenized_inputs = self.tokenizer(
                inputs, truncation=True,
                max_length=self.max_length,
                padding="longest", return_tensors="pt"
            ).to(self.device)
            
            class_input_ids = []
            class_attention_mask = []
            
            for labels_set in labels:
                tokenized_labels = self.labels_tokenizer(
                    labels_set, truncation=True,
                    max_length=self.max_length,
                    padding="max_length",
                    return_tensors="pt"
                ).to(self.device)
                class_input_ids.append(tokenized_labels["input_ids"])
                class_attention_mask.append(tokenized_labels["attention_mask"])
            
            tokenized_inputs["class_input_ids"] = torch.stack(class_input_ids)
            tokenized_inputs["class_attention_mask"] = torch.stack(class_attention_mask)

            labels_mask = [[1 for _ in range(len(labels[j]))] for j in range(len(texts))]
            tokenized_inputs["labels_mask"] = torch.tensor(labels_mask).to(self.device)
        return tokenized_inputs


class ZeroShotClassificationPipeline:
    """
    Main pipeline class for zero-shot classification with GLiClass models.
    
    Supports:
    - Hierarchical labels with dot notation (e.g., {"sentiment": ["positive", "negative"]})
    - Few-shot examples with <<EXAMPLE>> token
    - Task description prompts
    - Hierarchical output format matching input structure
    
    Example usage:
    
    ```python
    from gliclass import ZeroShotClassificationPipeline
    
    pipeline = ZeroShotClassificationPipeline(model, tokenizer)
    
    # === Hierarchical Labels for Review Classification ===
    hierarchical_labels = {
        "sentiment": ["positive", "negative", "neutral"],
        "topic": ["product", "service", "shipping"]
    }
    
    # Basic classification
    results = pipeline(
        "The product quality is amazing but delivery was slow",
        hierarchical_labels
    )
    # Results: [
    #     {'label': 'sentiment.positive', 'score': 0.89},
    #     {'label': 'topic.product', 'score': 0.92},
    #     {'label': 'topic.shipping', 'score': 0.76}
    # ]
    
    # === With Task Description Prompt ===
    results = pipeline(
        "The product quality is amazing but delivery was slow",
        hierarchical_labels,
        prompt="Classify this customer review by sentiment and topic:"
    )
    
    # === With Few-Shot Examples (uses <<EXAMPLE>> token) ===
    examples = [
        {
            "text": "Love this item, great quality!",
            "labels": ["sentiment.positive", "topic.product"]
        },
        {
            "text": "Customer support was unhelpful and rude",
            "labels": ["sentiment.negative", "topic.service"]
        },
        {
            "text": "Package arrived damaged after 2 weeks",
            "labels": ["sentiment.negative", "topic.shipping"]
        }
    ]
    
    results = pipeline(
        "Fast delivery and the item works perfectly!",
        hierarchical_labels,
        examples=examples,
        prompt="Classify customer feedback:"
    )
    
    # === Hierarchical Output (matches input structure) ===
    results = pipeline(
        "The product quality is amazing but delivery was slow",
        hierarchical_labels,
        return_hierarchical=True
    )
    # Returns:
    # {
    #     "sentiment": {
    #         "positive": 0.89,
    #         "negative": 0.05,
    #         "neutral": 0.12
    #     },
    #     "topic": {
    #         "product": 0.92,
    #         "service": 0.15,
    #         "shipping": 0.76
    #     }
    # }
    
    # === Per-Text Prompts ===
    results = pipeline(
        ["Electronics review text", "Clothing review text"],
        hierarchical_labels,
        prompt=["Analyze this electronics review:", "Analyze this clothing review:"]
    )
    ```
    """
    
    def __init__(
        self, 
        model, 
        tokenizer, 
        max_classes: int = 25, 
        max_length: int = 1024, 
        classification_type: str = 'multi-label', 
        device: str = 'cuda:0', 
        progress_bar: bool = True,
        label_separator: str = "."
    ):
        """
        Initialize the classification pipeline.
        
        Args:
            model: GLiClass model or path to model
            tokenizer: Tokenizer or path to tokenizer
            max_classes: Maximum number of classes to process
            max_length: Maximum sequence length
            classification_type: 'single-label' or 'multi-label'
            device: Device to run inference on
            progress_bar: Whether to show progress bar
            label_separator: Separator for hierarchical label notation (default: ".")
        """
        if isinstance(model, str):
            model = GLiClassBiEncoder.from_pretrained(model)
            
        self.label_separator = label_separator
            
        if model.config.architecture_type == 'uni-encoder':
            self.pipe = UniEncoderZeroShotClassificationPipeline(
                model, tokenizer, max_classes, max_length, 
                classification_type, device, progress_bar, label_separator
            )
        elif model.config.architecture_type in {'encoder-decoder', 'encoder-decoder-cls'}:
            self.pipe = EncoderDecoderZeroShotClassificationPipeline(
                model, tokenizer, max_classes, max_length, 
                classification_type, device, progress_bar, label_separator
            )
        elif model.config.architecture_type in {'bi-encoder', 'bi-encoder-fused'}:
            self.pipe = BiEncoderZeroShotClassificationPipeline(
                model, tokenizer, max_classes, max_length, 
                classification_type, device, progress_bar, label_separator
            )
        else:
            raise NotImplementedError("This architecture is not implemented")
    
    def flatten_labels(
        self, 
        labels: Union[List[str], Dict[str, Any]]
    ) -> List[str]:
        """
        Flatten hierarchical labels to dot notation.
        
        Example:
            >>> pipeline.flatten_labels({
            ...     "sentiment": ["positive", "negative", "neutral"],
            ...     "topic": ["product", "service", "shipping"]
            ... })
            ["sentiment.positive", "sentiment.negative", "sentiment.neutral",
             "topic.product", "topic.service", "topic.shipping"]
        """
        return flatten_hierarchical_labels(labels, separator=self.label_separator)
    
    def get_embeddings(self, *args, **kwargs):
        """Get embeddings for texts and labels."""
        return self.pipe.get_embeddings(*args, **kwargs)
    
    def __call__(
        self, 
        texts: Union[str, List[str]], 
        labels: Union[List[str], Dict[str, Any], List[List[str]], List[Dict[str, Any]]],
        threshold: float = 0.5, 
        batch_size: int = 8, 
        rac_examples: Optional[List] = None,
        examples: Optional[List[Dict[str, Any]]] = None,
        prompt: Optional[Union[str, List[str]]] = None,
        return_hierarchical: bool = False
    ):
        """
        Perform zero-shot classification.
        
        Args:
            texts: Single text or list of texts to classify
            labels: Labels - flat list or hierarchical dict
                Examples:
                - ["positive", "negative"] - flat labels
                - {"sentiment": ["positive", "negative"], "topic": ["product", "service"]}
            threshold: Classification threshold for multi-label (default: 0.5)
            batch_size: Batch size for processing
            rac_examples: Retrieval augmented examples (legacy)
            examples: Few-shot examples, each with 'text' and 'labels' keys
            prompt: Task description - string or list of strings (per-text)
            return_hierarchical: If True, return structure matching input labels
            
        Returns:
            List of predictions (flat) or hierarchical dicts with all scores
        """
        return self.pipe(
            texts, labels, 
            threshold=threshold, 
            batch_size=batch_size, 
            rac_examples=rac_examples,
            examples=examples,
            prompt=prompt,
            return_hierarchical=return_hierarchical
        )


class ZeroShotClassificationWithChunkingPipeline(BaseZeroShotClassificationPipeline):
    """Pipeline with long text chunking support."""
    
    def __init__(
        self, 
        model, 
        tokenizer, 
        max_classes: int = 25, 
        max_length: int = 1024, 
        classification_type: str = 'multi-label', 
        device: str = 'cuda:0', 
        progress_bar: bool = True,
        text_chunk_size: int = 8192, 
        text_chunk_overlap: int = 256, 
        labels_chunk_size: int = 8,
        label_separator: str = "."
    ):
        if isinstance(model, str):
            model = GLiClassModel.from_pretrained(model)
        super().__init__(
            model, tokenizer, max_classes, max_length, 
            classification_type, device, progress_bar, label_separator
        )
        
        self.text_chunk_size = text_chunk_size
        self.text_chunk_overlap = text_chunk_overlap
        self.labels_chunk_size = labels_chunk_size

    def chunk_text(self, text, chunk_size=None, overlap=None):
        """Split text into overlapping chunks."""
        if chunk_size is None:
            chunk_size = self.text_chunk_size
        if overlap is None:
            overlap = self.text_chunk_overlap
            
        if len(text) <= chunk_size:
            return [text]
        
        chunks = []
        start = 0
        while start < len(text):
            end = start + chunk_size
            chunk = text[start:end]
            chunks.append(chunk)
            
            if end >= len(text):
                break
                
            start = end - overlap
            
        return chunks

    def prepare_input(self, text, labels, examples=None, prompt=None):
        """
        Prepare input matching training format from data_processing.py:
        Order: Labels → SEP → Prompt → Text → Examples
        """
        input_parts = []

        # 1. Add labels
        for label in labels:
            label_tag = f"{self.label_token}{label}"
            input_parts.append(label_tag)
        input_parts.append(self.sep_token)

        # 2. Add task description prompt
        if prompt:
            input_parts.append(prompt)

        # 3. Format examples to go after text
        examples_str = ""
        if examples:
            examples_str = self._format_examples_for_input(examples)

        if self.model.config.prompt_first:
            return ''.join(input_parts) + text + examples_str
        else:
            return text + ''.join(input_parts) + examples_str

    def prepare_inputs(self, texts, labels, same_labels=False, examples=None, prompt=None):
        inputs = []

        if same_labels:
            for i, text in enumerate(texts):
                text_examples = None
                if examples:
                    if isinstance(examples[0], list):
                        text_examples = examples[i] if i < len(examples) else None
                    else:
                        text_examples = examples
                text_prompt = self._format_prompt(prompt, i)
                inputs.append(self.prepare_input(text, labels, text_examples, text_prompt))
        else:
            for i, (text, labels_) in enumerate(zip(texts, labels)):
                text_examples = None
                if examples:
                    if isinstance(examples[0], list):
                        text_examples = examples[i] if i < len(examples) else None
                    else:
                        text_examples = examples
                text_prompt = self._format_prompt(prompt, i)
                inputs.append(self.prepare_input(text, labels_, text_examples, text_prompt))

        tokenized_inputs = self.tokenizer(
            inputs, truncation=True,
            max_length=self.max_length,
            padding="longest", return_tensors="pt"
        ).to(self.device)
        return tokenized_inputs

    def aggregate_chunk_scores(
        self, 
        chunk_scores: List[Dict[str, float]], 
        labels: List[str]
    ) -> Dict[str, float]:
        """Aggregate scores across text chunks using max pooling."""
        aggregated = {label: 0.0 for label in labels}
        
        for scores in chunk_scores:
            for label, score in scores.items():
                aggregated[label] = max(aggregated[label], score)
                
        return aggregated

    @torch.no_grad()
    def process_single_text(self, text, labels, threshold=0.5, examples=None, prompt=None):
        """Process a single long text through chunks."""
        text_chunks = self.chunk_text(text)
        
        all_chunk_scores = []
        
        for text_chunk in text_chunks:
            chunk_logits = []
            all_labels = []
            
            for labels_idx in range(0, len(labels), self.labels_chunk_size):
                curr_labels = labels[labels_idx:labels_idx + self.labels_chunk_size]
                if labels_idx == 0:
                    all_labels = []
                all_labels.extend(curr_labels)
                
                tokenized_inputs = self.prepare_inputs(
                    [text_chunk], curr_labels, same_labels=True, 
                    examples=examples, prompt=prompt
                )
                model_output = self.model(**tokenized_inputs)
                logits = model_output.logits
                
                chunk_logits.extend(logits[0][:len(curr_labels)].tolist())
            
            text_logits = torch.tensor(chunk_logits)
            
            if self.classification_type == 'single-label':
                scores = torch.softmax(text_logits, dim=-1)
            else:
                scores = torch.sigmoid(text_logits)
            
            chunk_score_dict = {
                label: scores[i].item() for i, label in enumerate(all_labels)
            }
            all_chunk_scores.append(chunk_score_dict)
        
        aggregated_scores = self.aggregate_chunk_scores(all_chunk_scores, labels)
        
        if self.classification_type == 'single-label':
            total = sum(aggregated_scores.values())
            if total > 0:
                aggregated_scores = {k: v / total for k, v in aggregated_scores.items()}
            
            best_label = max(aggregated_scores, key=aggregated_scores.get)
            return [{'label': best_label, 'score': aggregated_scores[best_label]}], aggregated_scores
        
        else:
            text_results = []
            for label, score in aggregated_scores.items():
                if score >= threshold:
                    text_results.append({'label': label, 'score': score})
            text_results.sort(key=lambda x: x['score'], reverse=True)
            return text_results, aggregated_scores

    @torch.no_grad()
    def __call__(
        self, 
        texts, 
        labels, 
        threshold=0.5, 
        batch_size=8,
        labels_chunk_size=None, 
        text_chunk_size=None, 
        text_chunk_overlap=None,
        rac_examples=None,
        examples=None,
        prompt=None,
        return_hierarchical: bool = False
    ):
        """Classification with chunking for long texts."""
        original_labels = labels
        
        if labels_chunk_size is not None:
            self.labels_chunk_size = labels_chunk_size
        if text_chunk_size is not None:
            self.text_chunk_size = text_chunk_size
        if text_chunk_overlap is not None:
            self.text_chunk_overlap = text_chunk_overlap

        if isinstance(texts, str):
            if rac_examples:
                texts = retrieval_augmented_text(texts, rac_examples)
            texts = [texts]
        else:
            if rac_examples:
                texts = [
                    retrieval_augmented_text(text, ex) 
                    for text, ex in zip(texts, rac_examples)
                ]

        labels = self._process_labels(labels)

        short_texts, short_indices = [], []
        long_texts, long_indices = [], []
        
        for i, text in enumerate(texts):
            if len(text) <= self.text_chunk_size:
                short_texts.append(text)
                short_indices.append(i)
            else:
                long_texts.append(text)
                long_indices.append(i)

        results = [None] * len(texts)
        all_scores_list = [None] * len(texts)

        if short_texts:
            iterable = range(0, len(short_texts), batch_size)
            if self.progress_bar:
                iterable = tqdm(iterable, desc="Processing short texts")

            for idx in iterable:
                batch_texts = short_texts[idx:idx + batch_size]
                batch_indices = short_indices[idx:idx + batch_size]

                all_logits = [[] for _ in range(len(batch_texts))]
                all_labels = []

                for labels_idx in range(0, len(labels), self.labels_chunk_size):
                    curr_labels = labels[labels_idx:labels_idx + self.labels_chunk_size]
                    if labels_idx == 0:
                        all_labels = []
                    all_labels.extend(curr_labels)

                    batch_prompt = self._get_batch_prompt(prompt, idx, len(batch_texts))
                    tokenized_inputs = self.prepare_inputs(
                        batch_texts, curr_labels, same_labels=True, 
                        examples=examples, prompt=batch_prompt
                    )
                    model_output = self.model(**tokenized_inputs)
                    logits = model_output.logits

                    for i in range(len(batch_texts)):
                        all_logits[i].extend(logits[i][:len(curr_labels)].tolist())

                for i, orig_idx in enumerate(batch_indices):
                    text_logits = torch.tensor(all_logits[i])

                    if self.classification_type == 'single-label':
                        score = torch.softmax(text_logits, dim=-1)
                        pred_idx = torch.argmax(score).item()
                        pred_label = all_labels[pred_idx]
                        results[orig_idx] = [{'label': pred_label, 'score': score[pred_idx].item()}]
                        all_scores_list[orig_idx] = {all_labels[j]: score[j].item() for j in range(len(all_labels))}

                    elif self.classification_type == 'multi-label':
                        probs = torch.sigmoid(text_logits)
                        text_results = []
                        all_scores_list[orig_idx] = {all_labels[j]: probs[j].item() for j in range(len(all_labels))}
                        for j, prob in enumerate(probs):
                            score_val = prob.item()
                            if score_val >= threshold:
                                text_results.append({'label': all_labels[j], 'score': score_val})
                        text_results.sort(key=lambda x: x['score'], reverse=True)
                        results[orig_idx] = text_results

        if long_texts:
            iterable = range(len(long_texts))
            if self.progress_bar:
                iterable = tqdm(iterable, desc="Processing long texts")

            for i in iterable:
                text = long_texts[i]
                orig_idx = long_indices[i]
                text_prompt = self._format_prompt(prompt, orig_idx)
                text_results, all_scores = self.process_single_text(
                    text, labels, threshold, examples=examples, prompt=text_prompt
                )
                results[orig_idx] = text_results
                all_scores_list[orig_idx] = all_scores

        if return_hierarchical:
            hierarchical_results = []
            for i, (result, all_scores) in enumerate(zip(results, all_scores_list)):
                hierarchical_results.append(
                    build_hierarchical_output(result, original_labels, self.label_separator, all_scores)
                )
            return hierarchical_results

        return results


# Utility functions

def parse_hierarchical_prediction(prediction: str, separator: str = ".") -> Dict[str, Any]:
    """Parse dot-notation prediction into hierarchy levels."""
    parts = prediction.split(separator)
    result = {'full': prediction}
    for i, part in enumerate(parts):
        result[f'level_{i}'] = part
    return result


def group_predictions_by_hierarchy(
    predictions: List[Dict[str, Any]], 
    separator: str = "."
) -> Dict[str, List[Dict[str, Any]]]:
    """Group predictions by top-level category."""
    grouped = {}
    for pred in predictions:
        label = pred['label']
        parts = label.split(separator)
        top_level = parts[0] if parts else label
        
        if top_level not in grouped:
            grouped[top_level] = []
        grouped[top_level].append(pred)
    
    for key in grouped:
        grouped[key].sort(key=lambda x: x['score'], reverse=True)
    
    return grouped


def get_best_per_category(
    predictions: List[Dict[str, Any]],
    separator: str = "."
) -> Dict[str, Dict[str, Any]]:
    """Get best prediction per top-level category."""
    grouped = group_predictions_by_hierarchy(predictions, separator)
    return {category: preds[0] for category, preds in grouped.items() if preds}


from transformers import AutoConfig
from transformers.configuration_utils import PretrainedConfig
from transformers.utils import logging
from transformers.models.auto import CONFIG_MAPPING

from .utils import is_module_available

IS_TURBOT5 = is_module_available('turbot5')

if IS_TURBOT5:
    from turbot5.model.config import T5Config
else:
    from transformers import T5Config


logger = logging.get_logger(__name__)


class GLiClassModelConfig(PretrainedConfig):
    model_type = "GLiClass"
    is_composition = True

    def __init__(
        self,
        encoder_config = None,
        encoder_model=None,
        label_model_config=None,
        label_model_name=None,
        class_token_index = -1,
        text_token_index = -1,
        example_token_index = -1,
        ignore_index=-100,
        hidden_size=None,
        projector_hidden_act="gelu",
        vocab_size=None,
        problem_type='single_label_classification',
        max_num_classes=25,
        use_lstm=False,
        initializer_range=0.03,
        scorer_type='simple',
        pooling_strategy='first',
        class_token_pooling="first",
        focal_loss_alpha=0.5,
        focal_loss_gamma=2,
        focal_loss_reduction=None,
        logit_scale_init_value=2.6592,
        normalize_features=False,
        extract_text_features=False,
        contrastive_loss_coef=0,
        architecture_type = 'uni-encoder',
        prompt_first = False,
        squeeze_layers = False,
        layer_wise = False,
        encoder_layer_id = -1,
        embed_class_token = True,
        dropout = 0.1,
        use_segment_embeddings = False,
        **kwargs,
    ):
        if isinstance(encoder_config, dict):
            encoder_config["model_type"] = (encoder_config["model_type"] 
                                                if "model_type" in encoder_config 
                                                else "deberta-v2")
            if encoder_config['model_type'] == 't5':
                encoder_config = T5Config(**encoder_config)
            else:
                encoder_config = CONFIG_MAPPING[encoder_config["model_type"]](**encoder_config)
        elif encoder_config is None:
            encoder_config = CONFIG_MAPPING["deberta-v2"]()

        self.encoder_config = encoder_config
        self.encoder_model_name = encoder_model

        if label_model_name is not None:
            if isinstance(label_model_config, dict):
                label_model_config["model_type"] = (label_model_config["model_type"] 
                                                    if "model_type" in label_model_config 
                                                    else "deberta-v2")
                label_model_config = CONFIG_MAPPING[label_model_config["model_type"]](**label_model_config)
            elif label_model_config is None:
                label_model_config = CONFIG_MAPPING["deberta-v2"]()

            self.label_model_config = label_model_config
        else:
            self.label_model_config = None
        self.label_model_name = label_model_name

        if hidden_size is None:
            self.hidden_size = self.encoder_config.hidden_size
        else:
            self.hidden_size = hidden_size

        if vocab_size is None:
            self.vocab_size = self.encoder_config.vocab_size
        else:
            self.vocab_size = vocab_size
        
        if class_token_index == -1:
            self.class_token_index = self.vocab_size
        else:
            self.class_token_index = class_token_index
        
        if text_token_index == -1:
            self.text_token_index = self.vocab_size+1
        else:
            self.text_token_index = text_token_index

        if example_token_index == -1:
            self.example_token_index = self.vocab_size+2
        else:
            self.example_token_index = example_token_index

        self.ignore_index = ignore_index
        self.projector_hidden_act = projector_hidden_act
        self.problem_type = problem_type
        self.max_num_classes = max_num_classes
        self.initializer_range=initializer_range
        self.scorer_type = scorer_type
        self.pooling_strategy=pooling_strategy
        self.class_token_pooling=class_token_pooling
        self.use_lstm = use_lstm
        self.focal_loss_alpha=focal_loss_alpha
        self.focal_loss_gamma=focal_loss_gamma
        self.focal_loss_reduction = focal_loss_reduction
        self.contrastive_loss_coef=contrastive_loss_coef
        self.logit_scale_init_value = logit_scale_init_value
        self.normalize_features=normalize_features
        self.extract_text_features = extract_text_features
        self.architecture_type = architecture_type
        self.prompt_first = prompt_first
        self.squeeze_layers = squeeze_layers
        self.layer_wise = layer_wise
        self.encoder_layer_id = encoder_layer_id
        self.embed_class_token = embed_class_token
        self.pad_token_id = self.encoder_config.pad_token_id
        self.dropout = dropout
        self.use_segment_embeddings = use_segment_embeddings
        super().__init__(**kwargs)




import random
import torch
import copy
from dataclasses import dataclass
from torch.nn.utils.rnn import pad_sequence
from torch.utils.data import Dataset

@dataclass
class AugmentationConfig:
    """Configuration for data augmentation."""
    enabled: bool = True
    
    # Probability for each augmentation type
    random_label_removal_prob: float = 0.15
    random_label_addition_prob: float = 0.10
    random_text_addition_prob: float = 0.05
    random_add_description_prob: float = 0.25
    random_add_synonyms_prob: float = 0.1
    random_add_examples_prob: float = 0.25
    max_num_examples: int = 5

class DataAugmenter:
    def __init__(self, config, examples, labels, label2description=None):
        self.config = config
        self.examples = examples
        self.labels = sorted(labels)
        self.max_examples = self.config.max_num_examples
        self.label2description = label2description or {}
    
    def remove_labels(self, true_labels, all_labels):
        if len(all_labels)<=1:
            return true_labels, all_labels
        k = random.randint(1, len(all_labels))
        all_labels = random.sample(all_labels, k=k)
        true_labels = [lbl for lbl in true_labels if lbl in all_labels]
        return true_labels, all_labels

    def add_random_labels(self, all_labels):
        if not self.labels:
            return all_labels
        num_add = len(all_labels) + 1
        k = random.randint(1, min(num_add, len(self.labels)))
        add_labels = random.sample(self.labels, k=k)
        all_labels.extend(add_labels)
        return all_labels
    
    def add_random_text(self, text, all_labels):
        if not self.examples:
            return text
        example = random.sample(self.examples, k=1)[0]
        curr_labels = example['all_labels']
        joint_labels = set(all_labels) & set(curr_labels)
        if len(joint_labels):
            return text
        else:
            if random.randint(0, 1):
                text = example['text'] + ' ' + text
            else:
                text = text + ' ' + example['text']
            return text
    
    def add_random_synonyms(self, all_labels):
        """Replace some labels with their synonyms if available."""
        if not self.label2description:
            return all_labels
        
        augmented_labels = []
        for label in all_labels:
            if label in self.label2description:
                label_info = self.label2description[label]
                synonyms = label_info.get('synonyms', [])
                
                if synonyms and random.random() < 0.5:
                    augmented_labels.append(random.choice(synonyms))
                else:
                    augmented_labels.append(label)
            else:
                augmented_labels.append(label)
        
        return augmented_labels
    
    def add_random_descriptions(self, item):
        """Add descriptions to labels in the text or metadata."""
        if not self.label2description or not item['all_labels']:
            return item
        
        max_labels = min(3, len(item['all_labels']))
        labels_to_describe = random.sample(
            item['all_labels'], 
            k=random.randint(1, max_labels)
        )
        
        descriptions = []
        for label in labels_to_describe:
            if label in self.label2description:
                label_info = self.label2description[label]
                desc_list = label_info.get('descriptions', [])
                if desc_list:
                    descriptions.append(f"{label}: {random.choice(desc_list)}")
        
        if descriptions:
            desc_text = ' '.join(descriptions)
            if random.random() < 0.5:
                item['text'] = desc_text + ' ' + item['text']
            else:
                item['text'] = item['text'] + ' ' + desc_text
        
        return item
    
    def add_random_examples(self, item):
        """Add example texts with similar labels."""
        if not item['all_labels']:
            return item
        
        candidate_examples = item.get("examples", [])

        item_label_set = set(item['all_labels'])

        if not candidate_examples:
            
            for example in self.examples:
                example_label_set = set(example['true_labels'])
                example_text = example['text']

                overlap = item_label_set & example_label_set
                
                # Only consider examples with at least one overlapping label
                if overlap:
                    candidate_examples.append({"text": example_text, "labels": list(example_label_set)})
        
        if not candidate_examples:
            return item
        
        # Sort by overlap and select top examples
        random.shuffle(candidate_examples)
        top_candidates = candidate_examples[:self.max_examples]

        num_examples = random.randint(1, min(2, len(top_candidates)))
        selected_examples = random.sample(top_candidates, k=num_examples)
        
        item['examples'] = selected_examples
        
        return item
    
    def augment(self, item):
        if not self.config.enabled:
            return item

        text = copy.deepcopy(item['text'])
        true_labels = copy.deepcopy(item['true_labels'])
        all_labels = copy.deepcopy(item['all_labels'])

        # Create augmented item
        aug_item = {
            'text': text,
            'true_labels': true_labels,
            'all_labels': all_labels
        }
        
        # Copy any additional fields
        for key in item:
            if key not in aug_item:
                aug_item[key] = copy.deepcopy(item[key])

        if random.random() < self.config.random_label_removal_prob:
            aug_item['true_labels'], aug_item['all_labels'] = self.remove_labels(
                aug_item['true_labels'], aug_item['all_labels']
            )
    
        if random.random() < self.config.random_label_addition_prob:
            aug_item['all_labels'] = self.add_random_labels(aug_item['all_labels'])

        if random.random() < self.config.random_text_addition_prob:
            aug_item['text'] = self.add_random_text(aug_item['text'], aug_item['all_labels'])
        
        if random.random() < self.config.random_add_synonyms_prob:
            aug_item['all_labels'] = self.add_random_synonyms(aug_item['all_labels'])
        
        if random.random() < self.config.random_add_description_prob:
            aug_item = self.add_random_descriptions(aug_item)
        
        if random.random() < self.config.random_add_examples_prob:
            aug_item = self.add_random_examples(aug_item)

        return aug_item
    
class GLiClassDataset(Dataset):
    def __init__(self, examples, tokenizer, augment_config,
                            label2description={},
                            max_length=512, 
                            problem_type='multi_label_classification', 
                            architecture_type = 'uni-encoder',
                            add_description=True,
                            prompt_first=False,
                            get_negatives = False,
                            max_labels = 50,
                            labels_tokenizer=None,
                            shuffle_labels = True):
        self.tokenizer = tokenizer
        self.labels_tokenizer = labels_tokenizer
        self.label2description = label2description
        self.augment_config = augment_config
        self.max_length = max_length
        self._data = examples
        self.add_description = add_description
        self.problem_type = problem_type
        self.architecture_type = architecture_type
        self.prompt_first = prompt_first
        self.dataset_labels = self.collect_dataset_labels()
        self.get_negatives = get_negatives
        self.max_labels = max_labels
        self.shuffle_labels = shuffle_labels
        
        self.sep_token = "<<SEP>>"
        self.label_token = "<<LABEL>>"
        self.example_token = "<<EXAMPLE>>"
        self.augmenter = DataAugmenter(augment_config, examples, self.dataset_labels, label2description)
        print('Total labels: ', len(self.dataset_labels))
    
    def get_diversity(self):
        return [item.get("_diversity", {}).get("overall_diversity", 0.5) for item in self.data]
    
    def collect_dataset_labels(self):
        dataset_labels = set()
        for example in self._data:
            dataset_labels.update(set(example['all_labels']))
        return dataset_labels
    
    def prepare_labels(self, example, label2idx, problem_type):
        if problem_type == 'single_label_classification':
            labels = label2idx[example['true_labels'][0]]
        elif problem_type == 'multi_label_classification':
            if isinstance(example['true_labels'], dict):
                labels = [example['true_labels'][label] if label in example['true_labels'] else 0. for label in example['all_labels']]
            else:
                labels = [1. if label in example['true_labels'] else 0. for label in example['all_labels']]
        else:
            raise NotImplementedError(f"{problem_type} is not implemented.")
        return torch.tensor(labels)

    def prepare_prompt(self, item, label_token_first=True):
        prompt_texts = []
        for label in item['all_labels']:
            if label_token_first:
                label_tag = f"{self.label_token}{str(label)}"
            else:
                label_tag = f"{str(label)}{self.label_token}"
            prompt_texts.append(label_tag)
        prompt_texts.append(self.sep_token)
        prompt = item.get('prompt', '')
        prompt_texts.append(prompt)
        return prompt_texts

    def format_examples(self, item):
        examples = item.get("examples", [])
        if not examples:
            return ""
        examples = random.sample(examples, k=random.randint(1, len(examples)))
        parts = []
        for example in examples:
            parts.append(self.example_token)
            parts.append(example.get("text", ""))
            parts.append(" \nLabels:\n ")
            parts.append(', '.join(example.get("labels", example.get("true_labels", []))))
        parts.append(self.sep_token)
        return ''.join(parts)
    
    def tokenize(self, texts):
        tokenized_inputs = self.tokenizer(texts, truncation=True, max_length=self.max_length, padding="longest")
        return tokenized_inputs

    def tokenize_labels(self, labels):
        tokenized_inputs = self.labels_tokenizer(labels, truncation=True, max_length=self.max_length, padding="longest")
        return tokenized_inputs
    
    def tokenize_and_prepare_labels_for_uniencoder(self, example):
        if self.shuffle_labels:
            random.shuffle(example['all_labels'])
        input_text = self.prepare_prompt(example)
        examples_text = self.format_examples(example)
        if self.prompt_first:
            input_text = ''.join(input_text) + str(example['text']) + examples_text
        else:
            input_text = str(example['text']) + ''.join(input_text) + examples_text
        label2idx = {label: idx for idx, label in enumerate(example['all_labels'])}

        tokenized_inputs = self.tokenize(input_text)
        tokenized_inputs['labels'] = self.prepare_labels(example, label2idx, self.problem_type)
        tokenized_inputs['labels_text'] =  example['all_labels']
        tokenized_inputs['input_texts'] = example['text']
        return tokenized_inputs

    def tokenize_and_prepare_labels_for_encoder_decoder(self, example):
        if self.shuffle_labels:
            random.shuffle(example['all_labels'])
        class_texts = self.prepare_prompt(example, label_token_first=True)
        class_texts = ''.join(class_texts)
        examples_text = self.format_examples(example)

        label2idx = {label: idx for idx, label in enumerate(example['all_labels'])}

        input_text = str(example['text']) + examples_text
        tokenized_inputs = self.tokenize(input_text)
        tokenized_classes = self.tokenize(class_texts)
        tokenized_inputs["class_input_ids"] = tokenized_classes["input_ids"]
        tokenized_inputs["class_attention_mask"] = tokenized_classes["attention_mask"]   
        tokenized_inputs['labels'] = self.prepare_labels(example, label2idx, self.problem_type)
        return tokenized_inputs

    def tokenize_and_prepare_labels_for_biencoder(self, example):
        if self.shuffle_labels:
            random.shuffle(example['all_labels'])
        def prepare_prompt(labels):
            prompt_texts = []
            for label in labels:
                label_tag = f"<<LABEL>>"
                prompt_texts.append(label_tag)
            prompt_texts.append('<<SEP>>')
            return ''.join(prompt_texts)

        input_text = example['text']
        class_texts = example['all_labels']

        if self.architecture_type == 'bi-encoder-fused':
            prompt = prepare_prompt(class_texts)
            if self.prompt_first:
                input_text = f"{prompt} {input_text}"
            else:
                input_text = f"{input_text} {prompt}"

        tokenized_inputs = self.tokenize(input_text)
        tokenized_classes = self.tokenize_labels(class_texts)

        tokenized_inputs["class_input_ids"] = torch.tensor(tokenized_classes["input_ids"])
        tokenized_inputs["class_attention_mask"] = torch.tensor(tokenized_classes["attention_mask"])

        label2idx = {label: idx for idx, label in enumerate(example['all_labels'])}

        tokenized_inputs['labels_mask'] = torch.ones(len(class_texts))
        tokenized_inputs['labels'] = self.prepare_labels(example, label2idx, self.problem_type)
        return tokenized_inputs

    def __len__(self):
        return len(self._data)

    def __getitem__(self, idx):
        example = self._data[idx]

        example = self.augmenter.augment(example)

        if self.architecture_type == 'uni-encoder':
            model_inputs = self.tokenize_and_prepare_labels_for_uniencoder(example)
        elif self.architecture_type in {'encoder-decoder', 'encoder-decoder-cls'}:
            model_inputs = self.tokenize_and_prepare_labels_for_encoder_decoder(example)
        elif self.architecture_type in {'bi-encoder', 'bi-encoder-fused'}:
            model_inputs = self.tokenize_and_prepare_labels_for_biencoder(example)
        else:
            raise NotImplementedError('This architecture type is not implemented.')
        return model_inputs
    

def pad_2d_tensor(key_data):
    """
    Pad a list of 2D tensors to have the same size along both dimensions.
    
    :param key_data: List of 2D tensors to pad.
    :return: Tensor of padded tensors stacked along a new batch dimension.
    """
    if not key_data:
        raise ValueError("The input list 'key_data' should not be empty.")

    # Determine the maximum size along both dimensions
    max_rows = max(tensor.shape[0] for tensor in key_data)
    max_cols = max(tensor.shape[1] for tensor in key_data)
    
    tensors = []

    for tensor in key_data:
        rows, cols = tensor.shape
        row_padding = max_rows - rows
        col_padding = max_cols - cols
        # Pad the tensor along both dimensions
        padded_tensor = torch.nn.functional.pad(tensor, (0, col_padding, 0, row_padding),
                                                                 mode='constant', value=0)
        tensors.append(padded_tensor)

    # Stack the tensors into a single tensor along a new batch dimension
    padded_tensors = torch.stack(tensors)

    return padded_tensors

class DataCollatorWithPadding:
    def __init__(self, device = 'cuda:0'):
        self.device = device

    def __call__(self, batch):
        keys = batch[0].keys()
        padded_batch = {key: [] for key in keys}
        
        for key in keys:
            key_data = [item[key] for item in batch]
            if isinstance(key_data[0], torch.Tensor):
                if  key_data[0].dim() == 1:
                    padded_batch[key] = pad_sequence(key_data, batch_first=True)
                elif key_data[0].dim() == 2: 
                    padded_batch[key] = pad_2d_tensor(key_data)
            elif isinstance(key_data[0], list):
                data_el = "string"
                if len(key_data[0]):
                    data_el = key_data[0][0]
                if isinstance(data_el, str):
                    padded_batch[key] = key_data
                else:
                    max_length = max(len(seq) for seq in key_data)
                    padded_batch[key] = torch.tensor([seq + [0] * (max_length - len(seq)) 
                                                        for seq in key_data])
            elif type(key_data[0]) in {int, float}:
                padded_batch[key] = torch.tensor(key_data)
            elif isinstance(key_data[0], str):
                padded_batch[key] = key_data
            else:
                raise TypeError(f"Unsupported data type: {type(key_data[0])}")
        
        return padded_batch


from typing import Optional, Tuple, Dict, List, Union, Any, Callable
from tqdm import tqdm
import numpy as np
import os
from dataclasses import dataclass, field
import torch
from torch import nn
from transformers.trainer import (
    is_sagemaker_mp_enabled,
    get_parameter_names,
)
import torch.nn.functional as F
from torch.utils.data import DataLoader, Dataset

import transformers
from transformers import ZeroShotClassificationPipeline as TransformersClassificationPipeline
from .utils import default_f1_reward, is_module_available
from .pipeline import ZeroShotClassificationPipeline

if is_module_available("apex"):
    from apex import amp
    _has_apex = True
else:
    _has_apex = False
    amp = None

ALL_LAYERNORM_LAYERS = [nn.LayerNorm, nn.RMSNorm]

class EWC:
    """Elastic Weight Consolidation for preventing catastrophic forgetting in GLiClass models."""
    
    def __init__(
        self,
        model: nn.Module,
        dataset: Dataset,
        data_collator: Optional[Any] = None,
        device: str = 'cpu',
        ewc_lambda: float = 100.0,
        batch_size: int = 8,
        num_samples: Optional[int] = None,
        fisher_estimation_method: str = 'empirical',
        normalize_fisher: bool = True
    ):
        """Initialize EWC.
        
        Args:
            model: The GLiClass model to apply EWC to
            dataset: Dataset from previous task to compute Fisher information
            data_collator: Data collator for batching (required for GLiClass)
            device: Device to use for computation
            ewc_lambda: Importance weight for EWC penalty (higher = more regularization)
            batch_size: Batch size for Fisher computation
            num_samples: Number of samples to use for Fisher estimation (None = use all)
            fisher_estimation_method: Method for Fisher estimation ('empirical' or 'diagonal')
            normalize_fisher: Whether to normalize Fisher information values
        """
        self.model = model
        self.device = device
        self.ewc_lambda = ewc_lambda
        self.batch_size = batch_size
        self.num_samples = num_samples
        self.fisher_estimation_method = fisher_estimation_method
        self.normalize_fisher = normalize_fisher
        self.data_collator = data_collator
        
        # Store old parameters (deep copy to avoid reference issues)
        self.old_params: Dict[str, torch.Tensor] = {}
        for name, param in model.named_parameters():
            if param.requires_grad:
                self.old_params[name] = param.data.clone().detach()
        
        # Compute Fisher information matrix
        self.fisher_info: Dict[str, torch.Tensor] = self._compute_fisher(dataset)
        
        # Optionally normalize Fisher information
        if self.normalize_fisher:
            self._normalize_fisher()
    
    def _compute_fisher(
        self,
        dataset: Dataset
    ) -> Dict[str, torch.Tensor]:
        """Compute diagonal Fisher information matrix.
        
        The Fisher information measures how sensitive the loss is to changes
        in each parameter. Parameters with high Fisher information are important
        for the previous task.
        
        Args:
            dataset: Dataset to compute Fisher information from
            
        Returns:
            Dictionary mapping parameter names to Fisher information tensors
        """
        # Initialize Fisher information to zeros
        fisher: Dict[str, torch.Tensor] = {}
        for name, param in self.model.named_parameters():
            if param.requires_grad:
                fisher[name] = torch.zeros_like(param, device=self.device)
        
        # Set model to evaluation mode for consistent behavior
        was_training = self.model.training
        self.model.eval()
        
        # Create dataloader
        if self.num_samples is not None and self.num_samples < len(dataset):
            # Subsample dataset for efficiency
            indices = torch.randperm(len(dataset))[:self.num_samples].tolist()
            subset = torch.utils.data.Subset(dataset, indices)
            loader = DataLoader(
                subset,
                batch_size=self.batch_size,
                shuffle=False,
                collate_fn=self.data_collator
            )
        else:
            loader = DataLoader(
                dataset,
                batch_size=self.batch_size,
                shuffle=False,
                collate_fn=self.data_collator
            )
        
        num_batches = len(loader)
        
        print(f"Computing Fisher information from {len(loader.dataset)} samples...")
        
        # Compute Fisher information
        for batch in tqdm(loader, desc="Computing Fisher"):
            self.model.zero_grad()
            
            # Prepare inputs - handle GLiClass specific fields
            if isinstance(batch, dict):
                # Remove non-tensor fields that GLiClass might have
                inputs = {k: v for k, v in batch.items() 
                         if k not in ['labels_text', 'input_texts']}
                
                # Move tensors to device
                for k, v in inputs.items():
                    if isinstance(v, torch.Tensor):
                        inputs[k] = v.to(self.device)
            else:
                inputs = batch
            
            try:
                # Forward pass
                outputs = self.model(**inputs)
                
                if self.fisher_estimation_method == 'empirical':
                    # Use the actual loss for empirical Fisher
                    if hasattr(outputs, 'loss') and outputs.loss is not None:
                        loss = outputs.loss
                    else:
                        # Compute loss manually if not provided
                        logits = outputs.logits
                        labels = inputs.get('labels')
                        if labels is not None:
                            # Handle multi-label classification
                            if self.model.config.problem_type == 'multi_label_classification':
                                loss = F.binary_cross_entropy_with_logits(
                                    logits.view(-1),
                                    labels.view(-1).float()
                                )
                            else:
                                loss = F.cross_entropy(
                                    logits.view(-1, logits.size(-1)),
                                    labels.view(-1)
                                )
                        else:
                            continue
                else:
                    # Diagonal Fisher: sample from model's predictive distribution
                    logits = outputs.logits
                    
                    if self.model.config.problem_type == 'multi_label_classification':
                        probs = torch.sigmoid(logits)
                        # Sample binary labels
                        sampled_labels = torch.bernoulli(probs)
                        loss = F.binary_cross_entropy_with_logits(
                            logits.view(-1),
                            sampled_labels.view(-1)
                        )
                    else:
                        probs = F.softmax(logits, dim=-1)
                        log_probs = F.log_softmax(logits, dim=-1)
                        sampled_labels = torch.multinomial(
                            probs.view(-1, probs.size(-1)), 1
                        ).squeeze(-1)
                        loss = F.nll_loss(
                            log_probs.view(-1, log_probs.size(-1)),
                            sampled_labels
                        )
                
                # Backward pass to compute gradients
                loss.backward()
                
                # Accumulate squared gradients (Fisher information)
                for name, param in self.model.named_parameters():
                    if param.requires_grad and param.grad is not None:
                        fisher[name] += param.grad.data ** 2 / num_batches
                        
            except Exception as e:
                print(f"Warning: Error computing Fisher for batch: {e}")
                continue
        
        # Restore model training mode
        if was_training:
            self.model.train()
        
        return fisher
    
    def _normalize_fisher(self):
        """Normalize Fisher information values to prevent numerical issues."""
        # Compute max Fisher value across all parameters
        max_fisher = 0.0
        for name, fisher_val in self.fisher_info.items():
            max_fisher = max(max_fisher, fisher_val.max().item())
        
        if max_fisher > 0:
            # Normalize by max value
            for name in self.fisher_info:
                self.fisher_info[name] = self.fisher_info[name] / max_fisher
    
    def ewc_loss(self, batch_size: Optional[int] = None) -> torch.Tensor:
        """Compute EWC penalty loss.
        
        The EWC loss penalizes changes to parameters that were important
        for the previous task (as measured by Fisher information).
        
        Args:
            batch_size: Batch size for normalization (optional)
            
        Returns:
            EWC penalty loss tensor
        """
        loss = torch.tensor(0.0, device=self.device)
        
        for name, param in self.model.named_parameters():
            if param.requires_grad and name in self.fisher_info:
                # EWC penalty: F_i * (theta_i - theta_i^*)^2
                param_diff = param - self.old_params[name].to(param.device)
                fisher = self.fisher_info[name].to(param.device)
                loss += (fisher * param_diff ** 2).sum()
        
        # Optionally normalize by batch size
        if batch_size is not None:
            loss = loss / batch_size
        
        return self.ewc_lambda * loss
    
    def get_importance_scores(self) -> Dict[str, float]:
        """Get importance scores for each parameter group.
        
        Returns:
            Dictionary mapping parameter names to average importance scores
        """
        scores = {}
        for name, fisher in self.fisher_info.items():
            scores[name] = fisher.mean().item()
        return scores
    
    def update_lambda(self, new_lambda: float):
        """Update the EWC lambda value.
        
        Args:
            new_lambda: New lambda value for EWC penalty
        """
        self.ewc_lambda = new_lambda
    
    def consolidate(
        self,
        dataset: Dataset,
        alpha: float = 0.5
    ):
        """Consolidate knowledge by updating Fisher information with new task.
        
        This allows for online EWC where multiple tasks are consolidated.
        
        Args:
            dataset: Dataset from new task
            alpha: Mixing coefficient (0 = keep old Fisher, 1 = use new Fisher only)
        """
        # Compute new Fisher information
        new_fisher = self._compute_fisher(dataset)
        
        # Mix old and new Fisher information
        for name in self.fisher_info:
            if name in new_fisher:
                self.fisher_info[name] = (
                    (1 - alpha) * self.fisher_info[name] + 
                    alpha * new_fisher[name]
                )
        
        # Update old parameters
        for name, param in self.model.named_parameters():
            if param.requires_grad:
                self.old_params[name] = param.data.clone().detach()
        
        # Re-normalize if needed
        if self.normalize_fisher:
            self._normalize_fisher()

@dataclass
class TrainingArguments(transformers.TrainingArguments):
    cache_dir: Optional[str] = field(default=None)
    optim: str = field(default="adamw_torch")
    others_lr: Optional[float] = None
    others_weight_decay: Optional[float] = 0.0

    use_ewc: bool = field(
        default=False,
        metadata={"help": "Whether to use Elastic Weight Consolidation (EWC) for continual learning."}
    )
    ewc_lambda: float = field(
        default=100.0,
        metadata={"help": "Lambda parameter for EWC penalty. Higher values = more regularization."}
    )
    ewc_fisher_samples: Optional[int] = field(
        default=None,
        metadata={"help": "Number of samples to use for Fisher information estimation. None = use all."}
    )
    ewc_normalize_fisher: bool = field(
        default=True,
        metadata={"help": "Whether to normalize Fisher information values."}
    )
    ewc_gamma: float = field(
        default=0.95,
        metadata={"help": "Decay factor for Online EWC."}
    )


class Trainer(transformers.Trainer):
    """Extended Trainer with EWC support for continual learning."""
    
    def __init__(
        self,
        ewc: Optional[EWC] = None,
        prev_dataset=None,
        *args,
        **kwargs
    ):
        """Initialize Trainer with optional EWC support.

        Args:
            ewc: Pre-initialized EWC object (optional)
            prev_dataset: Previous dataset for EWC initialization (optional)
            *args: Arguments passed to parent Trainer
            **kwargs: Keyword arguments passed to parent Trainer
        """
        super().__init__(*args, **kwargs)

        # Ensure use_apex is set for compatibility with different transformers versions
        if not hasattr(self, 'use_apex'):
            self.use_apex = False

        self.ewc = ewc
        self.prev_dataset = prev_dataset
        self._ewc_initialized = ewc is not None
    
    def _maybe_initialize_ewc(self):
        """Initialize EWC if needed and not already initialized."""
        if self._ewc_initialized or not self.args.use_ewc:
            return
        
        if self.prev_dataset is None:
            print("Warning: EWC is enabled but no previous dataset provided. Skipping EWC initialization.")
            return
        
        print(f"Initializing EWC with lambda={self.args.ewc_lambda}...")
        
        # Get the data collator
        data_collator = self.data_collator
        
        # Determine device
        device = self.model.device if hasattr(self.model, 'device') else (
            next(self.model.parameters()).device if list(self.model.parameters()) else 'cpu'
        )
        
        # Create EWC instance
        ewc_kwargs = {
            'model': self.model,
            'dataset': self.prev_dataset,
            'data_collator': data_collator,
            'device': str(device),
            'ewc_lambda': self.args.ewc_lambda,
            'num_samples': self.args.ewc_fisher_samples,
            'normalize_fisher': self.args.ewc_normalize_fisher,
        }
        
        self.ewc = EWC(**ewc_kwargs)
        self._ewc_initialized = True
        print("EWC initialization complete.")

    def compute_loss(self, model, inputs, return_outputs=False, **kwargs):
        """Compute loss with optional EWC penalty.
        
        Args:
            model: The model
            inputs: Input batch
            return_outputs: Whether to return model outputs
            **kwargs: Additional arguments
            
        Returns:
            Loss tensor, or tuple of (loss, outputs) if return_outputs=True
        """
        # Get base loss from parent
        if return_outputs:
            loss, outputs = super().compute_loss(model, inputs, return_outputs=True, **kwargs)
        else:
            loss = super().compute_loss(model, inputs, return_outputs=False, **kwargs)
            outputs = None
        
        # Add EWC penalty if enabled
        if self.ewc is not None and self.args.use_ewc:
            batch_size = inputs.get('input_ids', inputs.get('labels')).shape[0] if isinstance(inputs, dict) else None
            ewc_loss = self.ewc.ewc_loss(batch_size=batch_size)
            loss = loss + ewc_loss
        
        if return_outputs:
            return loss, outputs
        return loss
    
    def train(self, *args, **kwargs):
        """Train with EWC initialization."""
        # Initialize EWC before training starts
        self._maybe_initialize_ewc()
        return super().train(*args, **kwargs)
    
    def training_step(self, model, inputs, *args, **kwargs) -> torch.Tensor:
        """
        Perform a training step on a batch of inputs.

        Subclass and override to inject custom behavior.

        Args:
            model (`nn.Module`):
                The model to train.
            inputs (`Dict[str, Union[torch.Tensor, Any]]`):
                The inputs and targets of the model.

                The dictionary will be unpacked before being fed to the model. Most models expect the targets under the
                argument `labels`. Check your model's documentation for all accepted arguments.

        Return:
            `torch.Tensor`: The tensor with training loss on this batch.
        """
        model.train()
        try:
            if "labels_text" in inputs:
                inputs.pop('labels_text')
            if "input_texts" in inputs:
                inputs.pop('input_texts')
            inputs = self._prepare_inputs(inputs)

            with self.compute_loss_context_manager():
                loss = self.compute_loss(model, inputs)

            del inputs
            torch.cuda.empty_cache()

            kwargs = {}

            if self.args.n_gpu > 1:
                loss = loss.mean()  # mean() to average on multi-gpu parallel training

            if self.use_apex and _has_apex:
                with amp.scale_loss(loss, self.optimizer) as scaled_loss:
                    scaled_loss.backward()
            else:
                self.accelerator.backward(loss, **kwargs)

            return loss.detach() / self.args.gradient_accumulation_steps
        except Exception as e:
            print(f"Skipping iteration due to error: {e}")
            model.zero_grad(set_to_none=True)
            torch.cuda.empty_cache()
            return torch.tensor(0.0, requires_grad=True).to(model.device)
        
    def prediction_step(
        self,
        model: torch.nn.Module,
        inputs: Dict[str, Union[torch.Tensor, Any]],
        prediction_loss_only: bool,
        ignore_keys: Optional[List[str]] = None,
    ) -> Tuple[Optional[torch.Tensor], Optional[torch.Tensor], Optional[torch.Tensor]]:
        """
        Perform an evaluation step on model using inputs.
        Subclass and override to inject custom behavior.
        Args:
            model (nn.Module):
                The model to evaluate.
            inputs (Dict[str, Union[torch.Tensor, Any]]):
                The inputs and targets of the model.
                The dictionary will be unpacked before being fed to the model. Most models expect the targets under the
                argument labels. Check your model's documentation for all accepted arguments.
            prediction_loss_only (bool):
                Whether or not to return the loss only.
            ignore_keys (List[str], *optional*):
                A list of keys in the output of your model (if it is a dictionary) that should be ignored when
                gathering predictions.
        Return:
            Tuple[Optional[torch.Tensor], Optional[torch.Tensor], Optional[torch.Tensor]]: A tuple with the loss,
            logits and labels (each being optional).
        """
        try:
            with torch.no_grad():
                if "labels_text" in inputs:
                    inputs.pop('labels_text')
                if "input_texts" in inputs:
                    inputs.pop('input_texts')
                loss = None
                with self.compute_loss_context_manager():
                    try:
                        outputs = model(**inputs)
                    except Exception as e:
                        raise RuntimeError(f"Error during model forward pass: {str(e)}")

                if not hasattr(outputs, 'loss'):
                    raise AttributeError("Model output does not contain 'loss' attribute")
                loss = outputs.loss

                if not hasattr(outputs, 'logits'):
                    raise AttributeError("Model output does not contain 'logits' attribute")
                logits = outputs.logits

                if 'labels' not in inputs:
                    raise KeyError("'labels' not found in input dictionary")
                labels = inputs['labels']

            if prediction_loss_only:
                return (loss, None, None)
            return (loss, logits, labels)

        except Exception as e:
            print(f"An error occurred during prediction step: {str(e)}")
            return (None, None, None)
        
    def create_optimizer(self):
        """
        Setup the optimizer.

        We provide a reasonable default that works well. If you want to use something else, you can pass a tuple in the
        Trainer's init through `optimizers`, or subclass and override this method in a subclass.
        """
        if is_sagemaker_mp_enabled():
            return super().create_optimizer()

        opt_model = self.model

        if self.optimizer is None:
            decay_parameters = get_parameter_names(opt_model, ALL_LAYERNORM_LAYERS)
            decay_parameters = [name for name in decay_parameters if "bias" not in name]
            if self.args.others_lr is not None:
                encoder_parameters = [name for name, _ in opt_model.named_parameters() if "encoder" in name]
                optimizer_grouped_parameters = [
                    {
                        "params": [
                            p for n, p in opt_model.named_parameters() if (n in decay_parameters and n not in encoder_parameters and p.requires_grad)
                        ],
                        "weight_decay": self.args.others_weight_decay,
                        "lr": self.args.others_lr,
                    },
                    {
                        "params": [
                            p for n, p in opt_model.named_parameters() if (n not in decay_parameters and n not in encoder_parameters and p.requires_grad)
                        ],
                        "weight_decay": 0.0,
                        "lr": self.args.others_lr,
                    },
                    {
                        "params": [
                            p for n, p in opt_model.named_parameters() if (n in decay_parameters and n in encoder_parameters and p.requires_grad)
                        ],
                        "weight_decay": self.args.weight_decay,
                    },
                    {
                        "params": [
                            p for n, p in opt_model.named_parameters() if (n not in decay_parameters and n in encoder_parameters and p.requires_grad)
                        ],
                        "weight_decay": 0.0,
                    },
                ]
            else:
                optimizer_grouped_parameters = [
                    {
                        "params": [
                            p for n, p in opt_model.named_parameters() if (n in decay_parameters and p.requires_grad)
                        ],
                        "weight_decay": self.args.weight_decay,
                    },
                    {
                        "params": [
                            p for n, p in opt_model.named_parameters() if (n not in decay_parameters and p.requires_grad)
                        ],
                        "weight_decay": 0.0,
                    },
                ]

            optimizer_cls, optimizer_kwargs = Trainer.get_optimizer_cls_and_kwargs(self.args)

            self.optimizer = optimizer_cls(optimizer_grouped_parameters, **optimizer_kwargs)

        return self.optimizer

@dataclass
class RLTrainerConfig(TrainingArguments):
    cliprange: float = field(
        default=0.2,
        metadata={"help": "Clip range."},
    )
    num_rl_iters: int = field(
        default=3,
        metadata={"help": "Number of RL iterations."},
    )
    gamma: float = field(
        default=-1,
        metadata={"help": "Focal loss gamma."},
    )
    alpha: float = field(
        default=-1,
        metadata={"help": "Focal loss alpha."},
    )
    labels_smoothing: float = field(
        default=-1,
        metadata={"help": "Labels smoothing factor."}
    )
    entropy_beta: float = field(
        default=-1,
        metadata={"help": "Coeficient of entropy factor."}
    )
    kl_beta: float = field(
        default=-1,
        metadata={"help": "Coeficient of KL-divergence factor."}
    )
    get_actions: str = field(
        default="bernoulli",
        metadata={"help": "How to get actions of a model, default is `bernoulli`, another option is `threshold`"},
    )
    threshold: float = field(
        default=0.5,
        metadata={"help": "Threshold value for predictions."},
    )

class RLTrainer(Trainer):
    def __init__(
        self,
        value_model: Optional[torch.nn.Module] = None,
        reference_model: Optional[Union[ZeroShotClassificationPipeline|TransformersClassificationPipeline]] = None,
        reward_components: Optional[List[Tuple[str, Callable]]] = None,
        *args,
        **kwargs
    ):
        super().__init__(*args, **kwargs)
        if value_model is not None:
            self.value_model = value_model.to(self.model.device)
        self.reference_model = reference_model
        if reward_components is None:
            reward_components = [('f1', default_f1_reward)]
        self.reward_components = reward_components
        self._init_metrics()

    def _init_metrics(self):
        self.metrics = {
            'total_loss': [],
            'advantages': [],
        }
        # Initialize metrics for each reward component
        for name, _ in self.reward_components.items():
            self.metrics[f'reward_{name}'] = []

    def compute_rewards(
        self,
        probs: torch.Tensor,
        actions: torch.Tensor,
        original_targets: torch.Tensor,
        valid_mask: torch.Tensor
    ) -> Dict[str, torch.Tensor]:
        rewards = {}
        total_reward = 0.0
        for name, reward_fn in self.reward_components.items():
            component = reward_fn(probs, actions, original_targets, valid_mask)
            rewards[name] = component
            total_reward += component
        rewards['total_reward'] = total_reward
        return rewards
    
    def get_reference_scores(self, input_texts, labels_text):
        if input_texts is None or labels_text is None:
            return None
        all_scores = []
        with torch.no_grad():
            if isinstance(self.reference_model, ZeroShotClassificationPipeline):
                results = self.reference_model(input_texts, labels_text, threshold=0.)
                for id, result in enumerate(results):
                    label2score = {item['label']: item['score'] for item in result}
                    label_scores = [label2score[label] for label in labels_text[id]]
                    all_scores.append(label_scores)
            elif isinstance(self.reference_model, TransformersClassificationPipeline):
                for text, labels in zip(input_texts, labels_text):
                    result = self.reference_model(text, labels)
                    label2score = {label:score for label, score in zip(result['labels'], result['scores'])}
                    label_scores = [label2score[label] for label in labels_text[id]]
                    all_scores.append(label_scores)
            else:
                raise NotImplementedError("This classification pipelines is not supported as a reference model.")
        max_length = max(len(seq) for seq in all_scores)
        all_scores = torch.FloatTensor([seq + [0] * (max_length - len(seq)) 
                                            for seq in all_scores]).to(self.model.device)
        return all_scores
    
    def compute_loss(
        self,
        inputs: torch.Tensor,
        targets: torch.Tensor,
        log_prob_prev: Optional[torch.Tensor] = None,
        value_outputs: Optional[torch.Tensor] = None,
        reference_probs: Optional[torch.Tensor] = None,
        **kwargs
    ) -> Tuple[torch.Tensor, torch.Tensor]:
        valid_mask = targets != -100
        original_targets = targets.clone()

        probs = torch.sigmoid(inputs)

        if self.args.get_actions == 'bernoulli':
            actions = torch.bernoulli(probs).detach()
        else:
            actions = (probs > self.args.threshold).float().detach()

        with torch.no_grad():
            metrics = self.compute_rewards(probs, actions, original_targets, valid_mask)

        reward = metrics['total_reward']

        if value_outputs is not None:
            state_values = value_outputs.logits[:, 0].unsqueeze(-1)  # Using first token logits as value prediction
            value_loss = torch.nn.functional.mse_loss(state_values, reward.detach())
        else:
            state_values = reward.mean()
            value_loss = torch.tensor(0.0).to(inputs.device)

        advantages = (reward - state_values).detach()
        self.metrics['advantages'].append(advantages.mean().item())

        for name, _ in self.reward_components.items():
            key = f'reward_{name}'
            self.metrics[key].append(metrics[name].mean().item())

        if self.args.label_smoothing_factor > 0:
            smoothed_actions = actions * (1 - self.args.label_smoothing_factor) + 0.5 * self.args.label_smoothing_factor
            log_prob_current = (
                smoothed_actions * torch.log(probs + 1e-8) +
                (1 - smoothed_actions) * torch.log(1 - probs + 1e-8)
            )
        else:
            log_prob_current = (
                actions * torch.log(probs + 1e-8) +
                (1 - actions) * torch.log(1 - probs + 1e-8)
            )

        if log_prob_prev is None:
            log_prob_prev = log_prob_current.detach()

        log_probs_diff = log_prob_current - log_prob_prev
        ratio = torch.exp(log_probs_diff)

        cliprange = self.args.cliprange
        per_label_loss1 = ratio * advantages
        per_label_loss2 = torch.clamp(ratio, 1 - cliprange, 1 + cliprange) * advantages
        loss_elements = -torch.min(per_label_loss1, per_label_loss2)

        loss_elements = loss_elements * valid_mask
        self.metrics['total_loss'].append(loss_elements.mean().item())

        if self.args.gamma > 0:
            p_t = probs * original_targets + (1 - probs) * (1 - original_targets)
            loss_elements = loss_elements * (p_t ** self.args.gamma)

        if self.args.alpha >= 0:
            alpha_t = self.args.alpha * original_targets + (1 - self.args.alpha) * (1 - original_targets)
            loss_elements = alpha_t * loss_elements

        loss = loss_elements.sum() / valid_mask.shape[0] + value_loss

        if reference_probs is not None:
            ref_per_token_logps = torch.log(reference_probs + 1e-8)
            per_token_logps = log_prob_current  
            per_label_kl = torch.exp(ref_per_token_logps - per_token_logps) - (ref_per_token_logps - per_token_logps) - 1
            per_label_kl = per_label_kl * valid_mask
            kl_loss = self.args.kl_beta * per_label_kl.mean()
            loss = loss + kl_loss

        if self.args.entropy_beta:
            entropy = - (probs * torch.log(probs + 1e-8) +
                        (1 - probs) * torch.log(1 - probs + 1e-8))
            loss = loss + self.args.entropy_beta * entropy.mean()

        return loss, log_prob_current


    def _inner_training_loop(self, *args, **kwargs):
        self.create_optimizer()
        if self.value_model is not None:
            value_optimizer = torch.optim.Adam(self.value_model.parameters(), lr=self.args.learning_rate)
        args = self.args
        accelerator = self.accelerator
        optimizer = self.optimizer
        model = self.model
        dataloader = self.get_train_dataloader()
        device = accelerator.device

        num_local_steps = len(dataloader)
        num_iters = args.num_train_epochs*num_local_steps
        pbar = tqdm(total=num_iters, desc="Training iterations")
        self._init_metrics()

        for epoch in range(args.num_train_epochs):
            self._init_metrics()
            model.train()
            if self.value_model is not None:
                self.value_model.train()

            for step, inputs in enumerate(dataloader):
                global_step = step+epoch*num_local_steps

                inputs = self._prepare_inputs(inputs)
                labels = inputs.pop('labels').to(device)
                if "labels_text" in inputs:
                    labels_text = inputs.pop('labels_text')
                else:
                    labels_text = None
                if "input_texts" in inputs:
                    input_texts = inputs.pop('input_texts')
                else:
                    input_texts = None
                prev_logps = None
                for iter in range(args.num_rl_iters):
                    try:
                        outputs = model(**inputs)
                        logits = outputs.logits
                        if self.value_model is not None:
                            value_outputs = self.value_model(**inputs)
                        else:
                            value_outputs = None
                        if self.reference_model is not None:
                            reference_probs = self.get_reference_scores(input_texts, labels_text)
                        else:
                            reference_probs = None
                        loss, current_logps = self.compute_loss(logits, labels, log_prob_prev=prev_logps, 
                                                                                value_outputs=value_outputs,
                                                                                reference_probs=reference_probs)
                    except Exception as e:
                        print(f"An error occurred during training step: {str(e)}")
                        del inputs
                        model.zero_grad(set_to_none=True)
                        torch.cuda.empty_cache()
                        break

                    accelerator.backward(loss)
                    if self.args.max_grad_norm is not None:
                        torch.nn.utils.clip_grad_norm_(model.parameters(), self.args.max_grad_norm)
                        if self.value_model is not None:
                            torch.nn.utils.clip_grad_norm_(self.value_model.parameters(), self.args.max_grad_norm)

                    optimizer.step()
                    optimizer.zero_grad()
                    if self.value_model is not None:
                        value_optimizer.step()
                        value_optimizer.zero_grad()

                    prev_logps = current_logps.detach()

                if global_step % args.logging_steps == 0:
                    self.log_metrics()

                if args.save_steps is not None and global_step % args.save_steps == 0:
                    self._save_checkpoint(model, step=global_step)

                pbar.set_postfix(epoch=epoch, step=step)
                pbar.update(1)

            if args.evaluation_strategy == "epoch":
                self.evaluate()

    def log_metrics(self):
        logged_metrics = {
            'loss': np.mean(self.metrics['total_loss']),
            'advantages': np.mean(self.metrics['advantages']),
        }
        # Add user reward components
        for name, _ in self.reward_components.items():
            key = f'reward_{name}'
            logged_metrics[key] = np.mean(self.metrics[key])
        self.log(logged_metrics)
        self._init_metrics()

    def _save_checkpoint(self, model, step=None):
        checkpoint_dir = f"checkpoint-{step}" if step else "final_model"
        output_dir = os.path.join(self.args.output_dir, checkpoint_dir)
        os.makedirs(output_dir, exist_ok=True)
        model.save_pretrained(output_dir)
        if self.tokenizer is not None:
            self.tokenizer.save_pretrained(output_dir)
        print(f"Checkpoint saved to {output_dir}")


import torch
import torch.nn.functional as F


def sequence_contrastive_loss(embeddings, mask):
    # embeddings shape: (B, L, D)
    # mask shape: (B, L)
    B, L, D = embeddings.shape

    # Normalize embeddings
    embeddings = F.normalize(embeddings, p=2, dim=-1)

    # Compute similarity matrix
    sim_matrix = torch.matmul(embeddings, embeddings.transpose(1, 2)) #/ self.temperature
    
    # Create labels for cross entropy (diagonal indices)
    labels = torch.arange(L, device=embeddings.device).unsqueeze(0).expand(B, -1)
    
    # Compute loss for each element in the batch
    loss = F.cross_entropy(sim_matrix.reshape(B*L, L), labels.reshape(-1), reduction='none')
    
    # Apply mask to loss
    loss = loss.view(B, L) * mask

    # Compute mean loss over non-padded elements
    loss = loss.sum() / mask.sum()

    return loss


def focal_loss_with_logits(
        inputs: torch.Tensor,
        targets: torch.Tensor,
        alpha: float = 0.25,
        gamma: float = 2,
        reduction: str = "none",
        label_smoothing: float = 0.0,
        ignore_index: int = -100  # default value for ignored index
) -> torch.Tensor:
    """
    Loss used in RetinaNet for dense detection: https://arxiv.org/abs/1708.02002.

    Args:
        inputs (Tensor): A float tensor of arbitrary shape.
                The predictions for each example.
        targets (Tensor): A float tensor with the same shape as inputs. Stores the binary
                classification label for each element in inputs
                (0 for the negative class and 1 for the positive class).
        alpha (float): Weighting factor in range (0,1) to balance
                positive vs negative examples or -1 for ignore. Default: ``0.25``.
        gamma (float): Exponent of the modulating factor (1 - p_t) to
                balance easy vs hard examples. Default: ``2``.
        reduction (string): ``'none'`` | ``'mean'`` | ``'sum'``
                ``'none'``: No reduction will be applied to the output.
                ``'mean'``: The output will be averaged.
                ``'sum'``: The output will be summed. Default: ``'none'``.
        label_smoothing (float): Specifies the amount of smoothing when computing the loss, 
                                                                where 0.0 means no smoothing.
        ignore_index (int): Specifies a target value that is ignored and does not contribute
                            to the input gradient. Default: ``-100``.
    Returns:
        Loss tensor with the reduction option applied.
    """
    # Create a mask to ignore specified index
    valid_mask = targets != ignore_index
    
    # Apply label smoothing if needed
    if label_smoothing != 0:
        with torch.no_grad():
            targets = targets * (1 - label_smoothing) + 0.5 * label_smoothing

    # Apply sigmoid activation to inputs
    p = torch.sigmoid(inputs)

    # Compute the binary cross-entropy loss without reduction
    loss = F.binary_cross_entropy_with_logits(inputs, targets, reduction="none")

    # Apply the valid mask to the loss
    loss = loss * valid_mask

    # Apply focal loss modulation if gamma is greater than 0
    if gamma > 0:
        p_t = p * targets + (1 - p) * (1 - targets)
        loss = loss * ((1 - p_t) ** gamma)

    # Apply alpha weighting if alpha is specified
    if alpha >= 0:
        alpha_t = alpha * targets + (1 - alpha) * (1 - targets)
        loss = alpha_t * loss

    # Apply reduction method
    if reduction == "none":
        return loss
    elif reduction == "mean":
        return loss.sum() / valid_mask.sum()  # Normalize by the number of valid (non-ignored) elements
    elif reduction == "sum":
        return loss.sum()
    else:
        raise ValueError(
            f"Invalid value for argument 'reduction': '{reduction}'. "
            f"Supported reduction modes: 'none', 'mean', 'sum'"
        )